import json
from datetime import datetime, timedelta
from math import pi
from typing import List

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import requests
from bokeh.embed import json_item
from bokeh.models import (ColumnDataSource, Grid, ImageURL, Legend, LinearAxis,
                          NumeralTickFormatter, Plot, Text, tickers)
from bokeh.palettes import brewer, d3
from bokeh.plotting import figure, show

from salmon.frontend.utils import get_logger, image_url

logger = get_logger(__name__)


def _make_hist(
    title,
    xlabel,
    hist,
    edges,
    width=600,
    height=200,
    toolbar_location="right",
    **kwargs,
):
    p = figure(
        title=title,
        background_fill_color="#fafafa",
        width=width,
        height=height,
        toolbar_location=toolbar_location,
        **kwargs,
    )
    p.quad(
        top=hist,
        bottom=0,
        left=edges[:-1],
        right=edges[1:],
        fill_color="blue",
        line_color="white",
        alpha=0.5,
    )

    p.y_range.start = 0
    p.legend.location = "center_right"
    p.legend.background_fill_color = "#fefefe"
    p.xaxis.axis_label = xlabel
    p.yaxis.axis_label = "Frequency"
    return p


async def _get_unique(series: pd.Series):
    assert series.nunique() == 1
    return series.iloc[0]


async def _get_nbins(x: np.array) -> int:
    if len(x) <= 10:
        return 10

    total_days = (np.nanmax(x) - np.nanmin(x)) / (60 * 60 * 24)
    bins = max(10, total_days * 4)
    return int(bins)


async def activity(df: pd.DataFrame, start_sec: float):
    x = df["time_received"].values.copy()
    bins = await _get_nbins(x)
    logger.info(f"bins = {bins}")
    bin_heights, edges = np.histogram(x, bins=bins)

    start = datetime(1970, 1, 1) + timedelta(seconds=start_sec)
    edges = [timedelta(seconds=e) + start for e in edges]

    _start = start.isoformat()[: 10 + 6]
    xlabel = f"\nTime received"
    p = _make_hist(
        f"Time responses received",
        xlabel,
        bin_heights,
        edges,
        toolbar_location=None,
        x_axis_type="datetime",
    )
    p.xaxis.major_label_orientation = pi / 4
    return p


async def _remove_outliers(x, low=True, high=True, frac=0.10):
    """Remove outliers ``frac`` fraction of the data"""
    p = (frac * 100) / 2
    _high = np.percentile(x, 100 - p)
    _low = np.percentile(x, p)
    if low and high:
        good = (x >= _low) & (x <= _high)
    elif low:
        good = x >= _low
    elif high:
        good = x <= _high
    return x[good]


async def response_time(df: pd.DataFrame):
    x = df["response_time"].values.copy()
    if len(x) >= 100:
        x = await _remove_outliers(x, low=False, high=True)
    bins = await _get_nbins(x)
    bin_heights, edges = np.histogram(x, bins=bins)
    p = _make_hist(
        f"Response time",
        "Time (s)",
        bin_heights,
        edges,
        width=300,
        toolbar_location=None,
    )
    return p


async def network_latency(df: pd.DataFrame):
    x = df["network_latency"].values.copy()
    if len(x) >= 100:
        x = await _remove_outliers(x, low=False, high=True)
    bins = await _get_nbins(x)
    bin_heights, edges = np.histogram(x, bins=bins)
    p = _make_hist(
        f"Time waiting for new query",
        "Time (s)",
        bin_heights,
        edges,
        width=300,
        toolbar_location=None,
    )
    return p


async def show_embedding(embedding: np.ndarray, targets: List[str], alg: str = ""):
    embedding = np.asarray(embedding)

    # scale each dimension between 0 and 1
    for dim in range(embedding.shape[1]):
        embedding[:, dim] = embedding[:, dim] - embedding[:, dim].min()
        embedding[:, dim] /= embedding[:, dim].max()
        embedding[:, dim] -= 0.5

    images = [
        k for k, target in enumerate(targets) if "img" in target or "video" in target
    ]
    image_urls = [image_url(target) for k, target in enumerate(targets) if k in images]

    data = {
        "x": embedding[images, 0],
        "y": embedding[images, 1] if len(embedding[0]) > 1 else embedding[images, 0],
        "image_url": image_urls,
    }
    source = ColumnDataSource(data=data)

    plot = figure(
        title=alg,
        plot_width=600,
        plot_height=500,
        toolbar_location="right",
        background_fill_color="#fafafa",
    )
    #  glyph = Text(x="x", y="y", text="text", angle=0.3, text_color="#96deb3")
    w = h = {"units": "data", "value": 0.1}
    w = h = {"units": "screen", "value": 80}
    glyph = ImageURL(x="x", y="y", url="image_url", w=w, h=h)
    plot.add_glyph(source, glyph)

    text = [k for k in range(len(targets)) if k not in images]
    data = {
        "x": embedding[text, 0],
        "y": embedding[text, 1] if len(embedding[0]) > 1 else embedding[text, 0],
        "text": [target for k, target in enumerate(targets) if k in text],
    }
    glyph = Text(x="x", y="y", text="text")
    source = ColumnDataSource(data=data)
    plot.add_glyph(source, glyph)
    return plot


async def _get_server_metrics():
    base = "http://prom:9090"
    start = datetime.now() - timedelta(days=1)
    end = datetime.now()
    data = {
        "query": "starlette_request_duration_seconds_bucket",
        "start": start.isoformat(),
        "end": end.isoformat(),
        "step": 0.1,
    }
    r = requests.post(base + "/api/v1/query", data=data)
    assert r.status_code == 200
    raw = r.json()
    assert raw["status"] == "success"
    rare = raw["data"]
    assert rare["resultType"] == "vector"
    med_rare = rare["result"]
    assert all(len(m["value"]) == 2 for m in med_rare)
    assert all(set(m.keys()) == {"value", "metric"} for m in med_rare)
    medium = [{"value": m["value"][1], **m["metric"]} for m in med_rare]
    df = pd.DataFrame(medium)
    df["value"] = df["value"].astype(float)
    df["le"] = df["le"].astype(float)

    cols = ["value", "le", "path"]
    proc = df[cols]
    proc.columns = ["count", "le", "endpoint"]

    bad_endpoints = [
        "/favicon.ico",
        "/metrics",
        "/api/v1/query",
        "/static",
        "/init_exp",
    ]
    idx = proc.endpoint.isin(bad_endpoints)
    idx |= proc.endpoint.isin([e + "/" for e in bad_endpoints])
    proc = proc[~idx].copy()
    return proc


async def _process_endpoint_times(p, endpoint):
    e = endpoint
    base = pd.DataFrame([{"count": 0, "le": 0, "endpoint": e}])
    p = p.append(base)
    p = p.sort_values(by="le")

    between = p["count"].diff()
    limits = p["le"].values
    upper = limits
    idx = np.arange(len(p))
    lower = limits[idx - 1]
    df = pd.DataFrame(
        {"between": between, "upper": upper, "lower": lower, "endpoint": e}
    )
    df["prob"] = df["between"] / df["between"].sum()
    return df.copy()


async def get_endpoint_time_plots():
    proc = await _get_server_metrics()
    endpoints = proc.endpoint.unique()
    dfs = {
        e: await _process_endpoint_times(proc[proc.endpoint == e], e) for e in endpoints
    }
    out = {}
    for e, df in dfs.items():
        logger.info(df.columns)
        x = [
            str(xi) if xi >= 0.1 or xi <= 0 else "{}ms".format(int(xi * 1000))
            for xi in df.upper.unique()
        ]

        p = figure(
            x_range=x,
            plot_height=150,
            toolbar_location=None,
            title=f"{e} processing time",
            width=500,
            tools="",
            background_fill_color="#fafafa",
        )

        _data = {k: df[k].values.tolist() for k in ["upper", "between"]}
        _data["upper"] = [str(k) for k in _data["upper"]]
        source = ColumnDataSource(_data)
        p.vbar(x=x, top=_data["between"], width=0.9, line_color="#" + "e" * 6)

        p.yaxis.axis_label = "Frequency"
        p.xaxis.axis_label = "Processing time (s)"
        p.yaxis.minor_tick_line_color = None  # turn off x-axis minor ticks

        hits = np.asarray(_data["between"])
        hits = hits[~np.isnan(hits)]
        if hits.sum() > 1:  # Only put plots in that have more than 1 hit
            out[e] = p
    return out


async def _get_alg_perf(df, agg="median"):
    cols = [c for c in df.columns if "time_" == c[:5] and c != "time_loop"]

    s = df[cols + ["time"]].copy()
    partial = df[list(cols)].copy().to_numpy()
    s["timedelta"] = pd.to_timedelta(s["time"] - s["time"].min(), unit="s")
    s = s.sort_values(by="timedelta")

    source = ColumnDataSource(s)

    names = {"_".join(v.split("_")[:-1]) for v in cols}
    lims = (partial.min(), partial.max())
    p = figure(
        title="Algorithm timing",
        x_axis_label="Time since start",
        y_axis_label="Time spent per task (s)",
        x_axis_type="datetime",
        width=600,
        height=300,
        toolbar_location="above",
        background_fill_color="#fafafa",
    )

    colors = d3["Category10"][10]
    for name, color in zip(names, colors):
        base = dict(x="timedelta", y=f"{name}_{agg}", source=source)
        p.line(**base, legend_label=name, line_width=2, line_color=color)
        p.circle(**base, size=5, color=color)
    p.legend.location = "top_left"
    return p


async def response_rate(df, n_sec=30):
    df = df.copy()
    df["time_since_start"] = df["time_received"] - df["time_received"].min()
    df["response_received"] = 1

    s = df[["response_received", "time_since_start"]].copy()
    s["time_since_start"] = s.time_since_start.apply(lambda x: timedelta(seconds=x))
    base = s.copy()

    for _sec in [1e-3, n_sec + 1e-3]:
        t = base.copy()
        t["time_since_start"] += timedelta(seconds=_sec)
        t["response_received"] = 0
        s = pd.concat((t, s))
    s = s.sort_values(by="time_since_start").set_index("time_since_start")

    ss = s.rolling(window=f"{n_sec}s").sum() / n_sec
    source = ColumnDataSource(ss)

    p = figure(
        title="Responses per second",
        x_axis_type="datetime",
        x_axis_label="Time since start",
        y_axis_label=f"({n_sec}s moving avg)",
        width=600,
        height=200,
        toolbar_location="above",
        background_fill_color="#fafafa",
    )
    p.varea(x="time_since_start", y1="response_received", source=source)
    return p


async def _get_query_db(df, agg="median"):
    d = df.copy()
    d["time_since_start"] = d["time"] - d["time"].min()
    d["datetime"] = d["time_since_start"].apply(
        lambda x: datetime(1970, 1, 1) + timedelta(seconds=x)
    )
    source = ColumnDataSource(d)

    Y = [c for c in d.columns if ("n_queries" in c) and (agg in c)]
    ratio = df[Y].max().max() / df[Y].min().min()
    kwargs = {} if ratio < 50 else dict(y_axis_type="log")
    p = figure(
        title="Database queries",
        x_axis_type="datetime",
        x_axis_label="Time since start",
        y_axis_label="Number of queries",
        width=600,
        height=200,
        toolbar_location="above",
        background_fill_color="#fafafa",
        **kwargs,
    )

    if not len(Y):
        logger.warning(f"No columns to plot! Y = {Y} but d.columns = {d.columns}")
        return None
    COLORS = d3["Category10"][len(Y)]
    lines = []
    for y, color in zip(Y, COLORS):
        base = dict(x="datetime", y=y, source=source)
        line = p.line(**base, line_color=color, line_width=2)
        p.circle(**base, size=5, color=color)
        lines.append([line])

    names = [y.replace("n_queries_", "").replace(f"_{agg}", "") for y in Y]
    items = list(zip(names, lines))
    legend = Legend(items=items, location="top_left")  # , label_width=130)
    p.add_layout(legend, "right")
    return p


async def _get_response_rate_plots(timestamps: pd.Series):
    """
    Parameters
    ----------
    timestamps : pd.Series
        Seconds responses received.
    """
    timestamps = timestamps.sort_values()
    timestamps -= timestamps.min()

    window = 1
    _rates_per_sec = (timestamps.copy() / window).astype(int).value_counts()
    rates_per_sec = _rates_per_sec.value_counts().sort_index()
    rates = rates_per_sec.index
    prob = rates_per_sec.to_numpy() / rates_per_sec.sum()
    rates = rates[prob >= 0.01]
    prob = prob[prob >= 0.01]

    rates = np.array(rates.tolist() + [rates.max() + 1])
    rates = rates - 1
    rates = (rates * 1.0) / window

    x = rates.copy()
    p1 = _make_hist(
        "Rate responses received",
        f"Rate (responses/sec over {window}s)",
        prob,
        x,
        width=300,
        toolbar_location=None,
        x_range=(x.min(), x.max()),
    )
    p1.xaxis.ticker = tickers.BasicTicker(min_interval=1)
    p1.xaxis[0].formatter = NumeralTickFormatter(format="0,0")
    p1.yaxis.axis_label = "Probability (empirical)"
    p1.yaxis[0].formatter = NumeralTickFormatter(format="0%")

    gaps = timestamps.diff().dropna()

    _bins = [[1 * 10 ** i, 2 * 10 ** i, 5 * 10 ** i] for i in range(-5, 5)]
    bins = [
        b
        for bins3 in _bins
        for b in bins3
        if np.percentile(gaps, 1) <= b <= np.percentile(gaps, 99)
    ]
    bin_heights, edges = np.histogram(gaps, bins=bins)
    p2 = _make_hist(
        f"Delay between responses",
        "Time (s)",
        bin_heights,
        edges,
        width=300,
        toolbar_location=None,
        x_axis_type="log",
    )

    meta = {
        "median_response_delay": "{:0.2f}".format(np.median(gaps)),
        "rate_mean": "{:0.2f}".format(_rates_per_sec.mean() / window),
        "rate_window": window,
    }
    return p1, p2, meta



from salmon._version import get_versions
__version__ = get_versions()["version"]
del get_versions

try:
    from salmon.backend import app as app_algs
    from salmon.frontend import app as app

    app.version = app_algs.version = __version__
except (ModuleNotFoundError, ImportError):
    pass

from salmon.triplets.offline import OfflineEmbedding



from salmon.triplets.samplers.adaptive._embed import (
    GD, OGD, CntsLRDamper, Embedding, GeoDamp, PadaDampG
)
from salmon.triplets.samplers.adaptive._noise_models import (
    CKL, GNMDS, SOE, STE, TSTE, Logistic
)
from salmon.triplets.samplers.adaptive._score import (
    InfoGainScorer, UncertaintyScorer
)



import numpy as np
import numpy.linalg as LA
import pytest
import scipy.spatial
import torch
from sklearn.utils import check_random_state

from salmon.triplets.samplers.adaptive.search import gram_utils


@pytest.mark.parametrize("n", [10, 20, 40])
@pytest.mark.parametrize("d", [2, 3])
def test_decompose_gram(n, d, seed=None):
    rng = check_random_state(seed)
    X = rng.randn(n, d)
    G = X @ X.T

    X_hat = gram_utils.decompose(G, d)

    assert not np.allclose(X, X_hat)
    Y1, Y2, m = scipy.spatial.procrustes(X, X_hat)
    assert m < 1e-15
    assert np.allclose(Y1, Y2)
    G_hat = X_hat @ X_hat.T
    assert np.allclose(G, G_hat)


@pytest.mark.parametrize("n", [10, 20, 40])
@pytest.mark.parametrize("d", [1, 2, 3])
def test_project_and_is_psd(n, d, seed=None):
    rng = check_random_state(seed)
    X = rng.randn(n, d)
    G = gram_utils.gram_matrix(X)
    assert gram_utils.is_psd(G)
    lamduhs, vecs = LA.eigh(G)
    lamduhs[0] = -1
    G = vecs.T @ np.diag(lamduhs) @ vecs
    assert not gram_utils.is_psd(G)

    F = gram_utils.onto_psd(G)
    assert not np.allclose(F, G)

    l2, _ = LA.eigh(F)
    assert np.allclose(min(l2), 0)


def test_project_changes_torch():
    n, d, seed = 20, 2, None
    X = torch.randn(n, d)

    G = gram_utils.gram_matrix(X.numpy())
    lamduhs, vecs = LA.eigh(G)
    lamduhs[0] = -1
    G = vecs.T @ np.diag(lamduhs) @ vecs
    G = torch.from_numpy(G)
    e, v = torch.linalg.eigh(G, UPLO="U")
    assert e.min().item() < -0.5

    before = G.numpy().copy()
    after = gram_utils.onto_psd(G.numpy(), out=G.numpy())

    assert not np.allclose(before, after)
    assert not torch.allclose(torch.from_numpy(before), G)
    e, v = torch.linalg.eigh(G, UPLO="U")
    assert e.min() > -0.25


def test_project_no_change_with_pos_eig(n=20, d=2, seed=None):
    rng = check_random_state(seed)
    X = rng.randn(n, d)
    G = gram_utils.gram_matrix(X)
    lamduhs, vecs = LA.eigh(G)
    lamduhs += lamduhs.min() + 1
    G = vecs.T @ np.diag(lamduhs) @ vecs

    F = gram_utils.onto_psd(G)
    assert np.allclose(F, G)


def test_gram_generation(seed=None, n=20, d=2):
    rng = check_random_state(seed)
    X = rng.randn(n, d)
    G_star = X @ X.T
    G_hat = gram_utils.gram_matrix(X)
    assert np.allclose(G_star, G_hat)
    assert np.allclose(G_hat, G_hat.T)

    for i, g_star in enumerate(G_star):
        for j, g_hat in enumerate(G_hat):
            assert np.allclose(G_hat[i, j], np.inner(X[i], X[j]))


def test_distances(seed=None, n=20, d=4):
    rng = check_random_state(seed)
    X = rng.randn(n, d)
    G = X @ X.T

    D = gram_utils.distances(G)

    D_star = scipy.spatial.distance.pdist(X) ** 2
    D_star = scipy.spatial.distance.squareform(D_star)
    assert np.allclose(D, D_star)


def test_dist2(seed=None):
    rng = check_random_state(seed)
    n, d = 10, 2
    X = rng.randn(n, d)
    G = X @ X.T

    inds = [rng.choice(n, size=2, replace=False) for _ in range(100)]
    dists = [LA.norm(X[i[0]] - X[i[1]]) ** 2 for i in inds]
    gram_dists1 = [gram_utils.dist2(G, i[0], i[1]) for i in inds]
    gram_dists2 = [gram_utils.dist2(G, i[1], i[0]) for i in inds]
    assert np.allclose(dists, gram_dists1)
    assert np.allclose(dists, gram_dists2)



from salmon.triplets.samplers._adaptive_runners import (
    ARR, CKL, GNMDS, SOE, SRR, STE, TSTE, Adaptive
)
from salmon.triplets.samplers._random_sampling import Random
from salmon.triplets.samplers._round_robin import RoundRobin
from salmon.triplets.samplers._test_runner import Test
from salmon.triplets.samplers._validation import Validation



import logging
import random
from collections import defaultdict
from copy import deepcopy
from datetime import datetime, timedelta
from textwrap import dedent
from typing import Any, Dict, List, Optional, Union

import numpy as np
from pydantic import BaseModel, BaseSettings, Field, validator

logger = logging.getLogger(__name__)


class Answer(BaseModel):
    """
    An answer to a triplet query. head, left and right are integers
    from '/get_query'. The 'winner' is an integer that is most similar to 'head',
    and must be one of 'left' and 'right'.

    'puid' is the "participant unique ID", and is optional.

    """

    head: int
    left: int
    right: int
    winner: int
    sampler: str
    score: float = 0
    puid: str = ""
    response_time: float = -1
    network_latency: float = -1


class Sampling(BaseSettings):
    """
    Settings to configure how more than two samplers are used.

    These settings are used in the HTML but are not stylistic. The
    exception is ``common``, which is passed to every ``sampler`` during
    initialization and will likely be optional arguments for
    :class:`~salmon.triplets.samplers.Adaptive`.

    The default configuration is specified completely below, which is rendered into a YAML file in `an example`_.

    .. _an example: https://github.com/stsievert/salmon/blob/master/examples/default.yaml


    """

    common: Dict[str, Any] = Field(
        {"d": 2},
        description="""Arguments to pass to every sampler for initialization (likely
        values for :class:`~salmon.triplets.samplers.Adaptive`; note that
        values for ``n`` and ``ident`` are already specified). Any
        values specified in this field will be overwritten by
        sampler-specific arguments.""",
    )
    probs: Optional[Dict[str, int]] = Field(
        None,
        description="""The percentage to sample each ``sampler`` when given the
        opportunity (which depends on ``samplers_per_user``). The percentages
        in this sampler must add up to 100.
        If not specified (default), choose each sampler with equal
        probability.""",
    )
    samplers_per_user: int = Field(
        0,
        description="""The number of samplers to assign to each user. Setting
        ``samplers_per_user=1`` means any user only sees queries generated
        from one sampler, and ``sampler_per_user=0`` means the user sees a
        new sampler every query""",
    )
    details: Dict[int, Any] = Field(
        {},
        description="""Different options for a deterministic choice of samplers.

        This dictionary is of the form ``{query_shown_to_user: options}``. The
        ``options`` is a dictionary with up to two keys:

        - ``sampler`` (required), which reflects which sampler receives the
          responses)

        - ``query`` (optional), which is a list of length 3 indicating the target
          indices appear in the query.

        For example, this YAML will ensure the 1st and 10th query the
        crowdsourcing user sees will be from:

        .. code-block:: yaml

          targets: [zero, one, two, three, four, five, six]
          # ^ list of (textual) targets; target "zero" has index 0 and
          # is targets[0] in Python

          samplers:
            ARR: {}
            Validation: {}
            valid2:
              class: Validation
              n_queries: 3

          sampling:
            probs: {ARR: 100, Validation: 0, valid2: 0}
            details:
              # Each key "n" is the n-th query the user sees
              # So here the 1st and 10th queries the user sees is customized
              1: {sampler: "Validation", query: [0, 2, 3]}
              10: {sampler: "valid2"}

          html:
            # ask 10 queries according to "sampling.probs".
            # The probabilistic sampling will be overriden by sampling.details
            max_queries: 10

        In this case, the crowdsourcing user will see the following:

        * 1st query shown will have head "zero", and feet "two" and "three".
        * Queries 2 and 9 will be generated by the :class:`~salmon.triplets.samplers.ARR` sampler.
        * The 10th query they see (also the last query): one of three
          (random) fixed/static queries.

        The sampler ``valid2`` will receive answers to 3 fixed/static queries,
        and the ``Validation`` sampler will receive answers to the query
        ``[0, 2, 3]``.
        """,
    )


class HTML(BaseSettings):
    """
    Stylistic settings to configure the HTML page.

    The default configuration is specified completely below, which is rendered into a YAML file in `an example`_.

    .. _an example: https://github.com/stsievert/salmon/blob/master/examples/default.yaml

    Arbitrary keys are allowed in this class
    (which might allow for customization on the HTML page).

    """

    class Config:
        extra = "allow"

    title: str = Field(
        "Similarity judgements",
        description="The title of the HTML page (the text shown in the tab bar)",
    )
    instructions: str = Field(
        "Please select the <i>comparison</i> item that is most similar to the <i>target</i> item.",
        description="The instructions the user sees above each query.",
    )
    debrief: str = Field(
        "<b>Thanks!</b> Please use the participant ID below.",
        description=dedent(
            """The message that the user sees after ``max_queries`` responses
            have been completed. The participant ID (``puid``) is shown below
            this message."""
        ),
    )
    skip_button: bool = Field(
        False,
        description="""Wheter to show a button to skip queries. Most
        useful when participants are instructed to skip queries only when the
        know nothing about any object.""",
    )
    css: str = Field(
        "",
        description="""CSS to be included the in the query page.  This CSS is
        inserted just before the ``</style>`` tag in the HTML head.""",
    )
    max_queries: int = Field(
        50,
        description="""The number of queries that the user will answer before
        seeing the ``debrief`` message). Set ``max_queries=0`` or
        ``max_queries=-1`` to ask unlimited queries.""",
    )
    arrow_keys: bool = Field(
        True,
        description="""Wheter to allow using the arrow keys as input.  Specifying
        ``arrow_keys=True`` might allow bad input (though there is a delay of
        200ms between queries).""",
    )
    query_params: List[str] = Field(
        ["puid"],
        description="""A list of flags to include in the params to /query.
        The default is ``query_params=['puid']``, which hits the endpoint
        ``/query?puid=foo`` if the participant's UID is ``foo``. Developers
        writing custom query pages might be interested in customizing this
        value.
        """,
    )

    js: str = Field(
        "",
        description="JavaScript to include in standalone ``<script>`` tag"
    )
    element_top: str = Field(
        "",
        description="Custom HTML to include at top of page"
    )
    element_middle: str = Field(
        "",
        description="Custom HTML to include in middle of page (between target and comparisons)"
    )
    element_bottom: str = Field(
        "",
        description="Custom HTML to include at top of page (above footer)"
    )
    element_standalone: str = Field(
        "",
        description="""Custom HTML to include as a standalone element, only
        encased in the ``head`` tag.
        """
    )



class Config(BaseSettings):
    """
    Customization of Salmon sampling and HTML.

    This class is often specified through uploading an ``init.yaml`` file as
    described in :ref:`yamlinitialization`.

    The default configuration is specified completely below, which is rendered into a YAML file in `an example`_.

    .. _an example: https://github.com/stsievert/salmon/blob/master/examples/default.yaml

    Some details on specific fields:

    * The only required field is targets.

        * ``targets`` is often specified with the upload of a ZIP file. See
          :ref:`yaml_plus_zip` for more detail. If not, specify a list of strings (which
          will be rendered as HTML).

    * ``sampling`` is not relevant until ``samplers`` is customized.

    Here's how to customize the config a bit with
    a YAML file:

    .. code-block:: yaml

       samplers:
         testing: {class: Random}
         ARR: {}
         Validation:
           n_queries: 20
       sampling:
         probs: {"ARR": 40, "Validation": 20, "testing": 20}
         common:
           d: 3  # dimensions to embed to; passed to every sampler
           random_state: 42  # argument to Adaptive
           initial_batch_size: 128  # argument to Embedding
       html:
         instructions: Click buttons, <b><i>or else.</i></b>

    Full documentation is below.

    .. note::

       ``sampler_per_user`` and ``detail`` only affect the HTML page
       shown to the user. They do not affect the backend code executed.
       They are not grouped with :class:`~salmon.triplets.manager.HTML`
       because they do affect the responses (at least when using
       Salmon's default config and not writing a custom query page).

    """

    targets: Optional[Union[int, List[str]]] = Field(
        None,
        description="""A list of targets that will be rendered as HTML. If a ZIP file is
            uploaded, this field is populated automatically.
            See :ref:`yaml_plus_zip` for more detail.""",
    )
    html: HTML = Field(
        HTML(),
        description="""Stylistic settings to configure the HTML page.
            See :class:`~salmon.triplets.manager.HTML` for more detail.""",
    )
    samplers: dict = Field(
        {"random": {"class": "Random"}},
        description="""Samplers to use, and their initialization parameters. See
            above or ":ref:`adaptive-config`" for more detail on
            customization, and :ref:`experiments` for experiments/benchmarks.""",
    )
    sampling: Sampling = Field(
        Sampling(),
        description="""Settings to configure how more than two samplers are
        used. See :class:`~salmon.triplets.manager.Sampling` for more detail.""",
    )

    def parse(self, user_config):
        self._update(user_config)
        self = self.parse_obj(user_config)
        self._validate()
        return self

    def _update(self, user_config):
        # Update user_config so when updated below, changes reflected.
        # (it's a plain dict, so it needs some help)
        if "sampling" in user_config and "common" in user_config["sampling"]:
            default_common = self.dict()["sampling"]["common"]
            default_common.update(user_config["sampling"]["common"])
            user_config["sampling"]["common"] = default_common

        self._warn(user_config)

    def _validate(self):
        if self.sampling.probs is None:
            # Sample each sampler equally
            n = len(self.samplers)
            freqs = [100 // n,] * n
            freqs[0] += 100 % n  # because integer division might be off by one
            sampling_percent = {k: f for k, f in zip(self.samplers, freqs)}
            self.sampling.probs = sampling_percent

        if set(self.sampling.probs.keys()) != set(self.samplers.keys()):
            sf = set(self.sampling.probs)
            s = set(self.samplers)
            msg = (
                "sampling.probs keys={} are not the same as samplers keys={}.\n\n"
                "Keys in sampling.probs but not in samplers: {}\n"
                "Keys in samplers but but in sampling.probs: {}\n\n"
            )
            raise ValueError(msg.format(sf, s, sf - s, s - sf))

        if (v := self.sampling.samplers_per_user) not in {0, 1}:
            raise NotImplementedError(
                "Only samplers_per_user in {0, 1} is implemented, not "
                f"samplers_per_user={v}"
            )

        if sum(self.sampling.probs.values()) != 100:
            msg = (
                "The values in sampling.probs should add up to 100; however, "
                "the passed sampling.probs={} adds up to {}"
            )
            s2 = self.sampling.probs
            raise ValueError(msg.format(s2, sum(s2.values())))

        if len(self.sampling.details):
            details = self.sampling.details
            if not all("sampler" in v for v in details.values()):
                bad_items = [v for v in details.items() if "sampler" not in v]
                msg = (
                    "Specify the key `sampler` for each element of "
                    "`sampling.details`. Received items {} without a key "
                    "'sampler'"
                )
                raise ValueError(msg.format(bad_items))
            details_samplers = set([v["sampler"] for v in details.values()])
            samplers = set(self.samplers.keys())
            if not details_samplers.issubset(samplers):
                msg = (
                    "Each sampler specified in sampling.details must be "
                    "created in samplers. In sampling.details, received  "
                    "key(s) {} which is not a subset of the samplers={}"
                )
                raise ValueError(msg.format(details_samplers, samplers))
            if any(not isinstance(d.get("query", []), list) for d in details.values()):
                bad_queries = {k: q.get("query", []) for k, q in details.items()}
                bad_queries = {
                    k: v for k, v in bad_queries.items() if not isinstance(v, list)
                }
                msg = (
                    "Not all `query` keys in sampling.details are lists. "
                    "Bad elements are {} "
                    "(show with {{num_queries_shown: supplied_query}}"
                )
                raise ValueError(msg.format(bad_queries))
            lengths = {len(d.get("query", [1, 2, 3])) for d in details.values()}
            if lengths != {3}:
                bad_queries = {k: q.get("query", [1, 2, 3]) for k, q in details.items()}
                bad_queries = {k: v for k, v in bad_queries.items() if len(v) != 3}
                msg = (
                    "Not all `query` keys in sampling.details have length 3. "
                    "Bad elements are {} "
                    "(show with {{num_queries_shown: supplied_query}}"
                )
                raise ValueError(msg.format(bad_queries))
            queries = {k: d.get("query", []) for k, d in details.items()}
            logger.warning("targets = %s", self.targets)
            n_targets = -1 + (
                len(self.targets) if isinstance(self.targets, list) else self.targets
            )
            bad_queries = {
                k: q
                for k, q in queries.items()
                if not (0 <= min(q) <= max(q) <= n_targets)
            }
            if len(bad_queries):
                msg = (
                    "Some target indices in the for sampling.details."
                    "query are too large. Shown in the form of"
                    "{{num_queries_shown: supplied_query}}, they are {}. "
                    "\n\n(reminder: here, Salmon/Python uses 0-based indexing, not 1-based indexing)"
                )
                raise ValueError(msg.format(bad_queries))

    def _warn(self, config):
        # TODO: deprecate
        html = self.html.dict()
        if any(h in config for h in html.keys()):
            misplaced_keys = [h for h in config if h in html]
            misplaced = [f"{h}: {config[h]}" for h in misplaced_keys]
            fmt_misplace = "\n  ".join(list(sorted(misplaced)))
            msg = (
                f"Move keys {misplaced_keys} into the `html` key. That is, include "
                f"this block of YAML:\n\nhtml:\n  {fmt_misplace}\n"
            )
            raise ValueError(msg)

        if "RandomSampling" in config.get("samplers", ""):
            raise ValueError(
                "The sampler `RandomSampling` has been renamed to `Random`."
            )


def deserialize_query(serialized_query: str) -> Dict[str, int]:
    h, l, r = serialized_query.split("-")
    flip = random.choice([True, False])
    if flip:
        l, r = r, l
    return {
        "head": int(h),
        "left": int(l),
        "right": int(r),
    }


def get_responses(answers: List[Dict[str, Any]], targets, start_time=0):
    start = start_time
    out = []

    user_responses = defaultdict(int)
    for datum in answers:
        out.append(datum)
        puid = datum["puid"]
        user_responses[puid] += 1
        datetime_received = timedelta(seconds=datum["time_received"]) + datetime(
            1970, 1, 1
        )
        idxs = {
            key + "_html": targets[datum[key]]
            for key in ["left", "right", "head", "winner", "loser"]
        }
        names = {
            key + "_filename": _get_filename(idxs[f"{key}_html"])
            for key in ["left", "right", "head", "winner", "loser"]
        }
        meta = {
            "time_received_since_start": datum["time_received"] - start,
            "datetime_received": datetime_received.isoformat(),
            "start_time": start_time,
            "puid_num_responses": user_responses[puid],
            "num_responses": sum(user_responses.values()),
        }
        out[-1].update({**idxs, **names, **meta})
    return out


def random_query(n: int) -> Dict[str, int]:
    rng = np.random.RandomState()
    a, b, c = rng.choice(n, size=3, replace=False)
    return {"head": int(a), "left": int(b), "right": int(c)}


def _get_filename(html):
    html = str(html)
    if "<img" in html or "<video" in html:
        i = html.find("src=")
        j = html[i:].find(" ")
        return html[i + 5 : i + j - 1].replace("/static/targets/", "")
    return html



import random

import numpy as np
import pandas as pd
import pytest

from .utils import server


@pytest.mark.parametrize("sampler", ["RoundRobin", "Random"])
def test_targets(sampler, server):
    samplers = {sampler: {"targets": [0, 2, 4]}, "Test": {"class": "Random"}}
    config = {"samplers": samplers, "targets": 5}
    server.authorize()
    server.post("/init_exp", data={"exp": config})

    for k in range(30):
        q = server.get("/query").json()
        ans = {"winner": random.choice([q["left"], q["right"]]), "puid": "foo", **q}
        server.post("/answer", json=ans)
    r = server.get("/responses")
    df = pd.DataFrame(r.json())

    allow = df.loc[df.sampler == sampler, ["left", "right", "head"]].to_numpy()
    test = df.loc[df.sampler == "Test", ["left", "right", "head"]].to_numpy()

    assert set(np.unique(allow)) == {0, 2, 4}
    assert set(np.unique(test)) == {0, 1, 2, 3, 4}


@pytest.mark.parametrize("sampler", ["RoundRobin", "Random"])
def test_targets_too_large(sampler, server):
    samplers = {sampler: {"targets": [0, 2, 5]}}
    config = {"samplers": samplers, "targets": 5}
    server.authorize()
    r = server.post("/init_exp", data={"exp": config}, error=True)
    assert r.status_code == 500
    assert "At least one targets target is too large" in r.text


@pytest.mark.parametrize("sampler", ["RoundRobin", "Random"])
def test_targets_wrong_type(sampler, server):
    samplers = {sampler: {"targets": 0}}
    config = {"samplers": samplers, "targets": 5}
    server.authorize()
    r = server.post("/init_exp", data={"exp": config}, error=True)
    assert r.status_code == 500
    assert "Specify a list for targets. Got" in r.text


@pytest.mark.parametrize("sampler", ["RoundRobin", "Random"])
def test_targets_not_integers(sampler, server):
    samplers = {sampler: {"targets": [0.1, 0.2, 0.3]}}
    config = {"samplers": samplers, "targets": 5}
    server.authorize()
    r = server.post("/init_exp", data={"exp": config}, error=True)
    assert r.status_code == 500
    assert "Not all items in targets are integers" in r.text


@pytest.mark.parametrize("sampler", ["RoundRobin", "Random"])
def test_targets_too_small(sampler, server):
    samplers = {sampler: {"targets": [0, 1]}}
    config = {"samplers": samplers, "targets": 5}
    server.authorize()
    r = server.post("/init_exp", data={"exp": config}, error=True)
    assert r.status_code == 500
    assert "Specify at least 3 targets items. Got" in r.text



"""
This file tests to make sure that all the examples launch without errors.

Goal: ensure the YAML files aren't out of date).
Goal: ensure the examples at least launch and run.
Non-goal: testing to make sure that the backend runs without errors.
    (we know backend init errors propogate because of
    test_backend.test_init_errors_propogate)

"""
import random
from pathlib import Path
from zipfile import ZipFile

import pytest
import yaml

from .utils import server

EG_DIR = Path(__file__).parent.parent / "examples"
SUBDIRS = [
    f.name for f in EG_DIR.iterdir() if f.is_dir() and f.name[0] not in ["_", "."]
]
YAMLS = [f.name for f in EG_DIR.glob("*.yaml")]


def test_defaults_config(server):
    server.authorize()
    targets = ["actually", "required", "with", "zip", "or", "yaml"]
    r = server.post("/init_exp", data={"exp": {"targets": targets}})
    assert r.status_code == 200
    defaults = server.get("/config").json()
    rendered = yaml.safe_load((EG_DIR / "default.yaml").read_text())
    assert defaults == rendered


def _test_upload(exp: Path, target_zip: Path = None, server=None):
    assert server is not None
    server.authorize()
    if target_zip:
        targets = Path(target_zip)
        assert targets.exists()
        t = targets.read_bytes()
        assert len(t) > 0
        assert t[:4] == b"\x50\x4B\x03\x04"

        kwargs = {"files": {"targets": t}}
    else:
        kwargs = {}
    r = server.post("/init_exp", data={"exp": exp.read_text()}, timeout=60, **kwargs)
    return r.status_code == 200


def _test_example(exp, target_zip=None, server=None):
    assert server is not None
    success = _test_upload(exp, target_zip, server)
    assert success
    server.get("/", timeout=5)
    for _ in range(10):
        r = server.get("/query")
        assert r.status_code == 200
        query = r.json()
        query["winner"] = random.choice([query["left"], query["right"]])
        r = server.post("/answer", data=query)
        assert r.status_code == 200
    server.reset()
    return True


@pytest.mark.parametrize("fname", YAMLS)
def test_basic_examples(fname: str, server):
    server.authorize()
    success = _test_example(EG_DIR / fname, server=server)
    assert success


@pytest.mark.parametrize("eg_dir", SUBDIRS)
def test_directory_examples(eg_dir: str, server):
    _eg_dir = EG_DIR / eg_dir
    server.authorize()
    for exp in _eg_dir.glob("*.yaml"):
        for target_zip in _eg_dir.glob("*.zip"):
            success = _test_example(exp, target_zip, server)
            assert success



import random
from time import sleep

import pytest

from .utils import LogError, logs, server


def test_init_w_queries_valid_idx(server, logs):
    samplers = {"Validation": {"queries": [[0, 1, 4], [0, 1, 5]]}}
    exp = {"targets": [0, 1, 2, 3], "samplers": samplers}

    with pytest.raises(LogError), logs:
        server.authorize()
        server.get("/init")
        r = server.post("/init_exp", data={"exp": exp}, error=True)
        assert r.status_code == 500
        assert "the index 5 is included" in r.text.lower()
        assert "the n=4 targets" in r.text.lower()


def test_validation_sampling(server, logs):
    n_val = 4
    exp = {
        "targets": list(range(10)),
        "samplers": {"Validation": {"n_queries": n_val}},
    }
    data = []
    puid = "adsfjkl4awjklra"

    n_repeat = 4
    server.authorize()
    server.post("/init_exp", data={"exp": exp})
    Q = []
    for k in range(n_repeat * n_val):
        q = server.get("/query").json()
        _ans = random.choice([q["left"], q["right"]])
        ans = {"winner": _ans, "puid": k, **q}
        print([ans[k] for k in ["head", "left", "right"]])
        Q.append(ans)
        server.post("/answer", data=ans)
        data.append(ans)
        sleep(0.20)

    # Test the number of unique queries is specified by n_val
    queries = [(q["head"], (q["left"], q["right"])) for q in Q]
    uniq_queries = [(h, min(c), max(c)) for h, c in queries]
    assert len(set(uniq_queries)) == n_val, set(uniq_queries)

    # Test the order gets shuffled every iteration
    order = [hash(q) for q in queries]
    round_orders = [order[k * n_val : (k + 1) * n_val] for k in range(n_repeat)]
    for round_order in round_orders:
        assert len(set(round_order)) == n_val
    same_order = [round_orders[0] != order for order in round_orders[1:]]
    assert sum(same_order) in [len(same_order), len(same_order) - 1]



from typing import Union

import numpy as np
import numpy.linalg as LA
import scipy
import torch
from scipy.linalg import eigh
from scipy.spatial import procrustes

Array = Union[np.ndarray, torch.Tensor]


def decompose(G, d):
    """
    Arguments
    ---------
    G : ndarray
        Gram matrix; X @ X.T
    d : int
        Dimension of each vector in X; X.shape == (n, d)
        when G.shape == (n, n)

    Returns
    -------
    X : ndarray
        Points that make up gram matrix
    """
    assert_gram(G)
    n = G.shape[0]
    w, v = eigh(G)
    i = [idx for idx in range(n - d, n)]
    assert len(i) == d
    X_hat = np.diag(np.sqrt(w[i])) @ (v[:, i]).T
    return X_hat.T


def gram_matrix(X: Array) -> Array:
    """
    Get Gram matrix from embedding

    Arguments
    ---------
    X : Array
        Embedding. X.shape == (num_items, item_dim)

    Returns
    -------
    G : Array
        Gram matrix. G.shape == (n, n)
    """
    if isinstance(X, torch.Tensor):
        return X @ X.transpose(0, 1)
    return X @ X.T


def distances(G: Array) -> Array:
    """
    Get distance matrix from gram matrix

    Arguments
    ---------
    G : Array
        Gram matrix. G.shape == (n, n) for n objects

    Returns
    -------
    D : Array
        Distance matrix. D.shape == (n, n)
    """
    assert_gram(G)
    G1 = np.diag(G).reshape(1, -1)
    G2 = np.diag(G).reshape(-1, 1)

    D = -2 * G + G1 + G2
    return D


def dist2(G, a, b):
    # assert_gram(G)
    return G[a, a] + G[b, b] - 2 * G[a, b]


def is_psd(G, return_vals=False):
    s = eigh(G, eigvals_only=True)
    psd = 0 <= s.min() or s.min() > -3e-7
    return psd if not return_vals else (psd, s)


def onto_psd(G, one=True, out=None):
    """
    Project onto semi-positive definite cone
    """
    # assert_gram(G)
    if out is None:
        out = G.copy()
    s, v = eigh(out, eigvals=(0, 0))
    assert s.shape == (1,)
    assert v.shape[1] == 1
    v = v.flat[:]
    if s < 0:
        out -= s * np.outer(v, v)
    return out


def assert_embedding(X):
    n, d = X.shape
    assert n != d


def assert_gram(G):
    pass


#     if isinstance(G, torch.Tensor):
#         assert torch.allclose(G, G.transpose(0, 1))
#         m = torch.abs(torch.diag(G) / torch.norm(G))
#         assert not np.allclose(m.sum().item(), 0)
#     else:
#         assert np.allclose(G, G.T)
#         assert not np.allclose(np.diag(G) / LA.norm(G), 0)


def assert_distance(D):
    pass


#     assert np.abs(np.diag(D)).sum() < 1e-6
#     assert np.allclose(D, D.T)



from time import time

import gram_utils
import numpy as np
import numpy.linalg as LA
import search
from gram_utils import dist2


def random_query(n):
    return np.random.choice(n, size=3, replace=False)


def update(G, head, winner, loser):
    """
    Perform the update mentioned in [1] (which also uses a Gram matrix)

    [1]:"Efficient online relative comparison kernel learning" by Eric Heim et. al.
    """
    D = gram_utils.distances(G)
    if D[head, winner] < D[head, loser]:
        pred_winner, pred_loser = winner, loser
    else:
        pred_winner, pred_loser = loser, winner

    # probability needs to use the estimated probability
    a = dist2(G, head, pred_winner)
    b = dist2(G, head, pred_loser)
    p_lose = 1 / (1 + np.exp(a - b))
    p_win = 1 - p_lose
    # 1 / (1 + exp(a - b)) = 1 / (1 + exp((-b) - (-a)))
    # = 1 / (1 + exp(-b) / exp(-a))
    # = exp(-a) / (exp(-a) + exp(-b))
    # and
    # a = ||h - w||^2_2
    # b = ||h - l||^2_2
    # p_win \probto exp(-||head - winner||^2) as in the STE paper

    # the amount of change should depend on the actual winner and loser
    # (and not change if it agrees)
    d2_win = dist2(G, head, winner)
    d2_loser = dist2(G, head, loser)
    P = np.log(np.e) / (1 + np.log(np.e))
    P = 0.8
    kappa = np.log(P) - np.log(1 - P)
    aggressive = d2_win - d2_loser + kappa
    step_size = max(0, aggressive)
    amount = step_size * (1 - p_win)

    # performing G = G - (amount * grad)
    # (grad given by 1/2 values)
    # - (x) given by G -= x
    G[head, winner] -= -2 * amount
    G[winner, head] -= -2 * amount
    G[head, loser] -= 2 * amount
    G[loser, head] -= 2 * amount
    G[winner, winner] -= 1 * amount
    G[loser, loser] -= -1 * amount

    start = time()
    G = gram_utils.project(G)
    # G_max_eig = LA.norm(G, ord=2)
    # if G_max_eig > 10:
    #    G *= 10 / G_max_eig
    svd_time = time() - start
    # if 10 < LA.norm(G):
    #   G /= LA.norm(G)

    # looking to see what vectors have 20 < norm(x)
    max_norm = 2
    too_large = np.argwhere(max_norm ** 2 < np.diag(G))
    norms = np.sqrt(G[too_large, too_large])

    # making sure those vectors x have norm(x) = 20
    for idx, norm in zip(too_large, norms):
        G[idx, :] *= max_norm / norm
        G[:, idx] *= max_norm / norm
    return G


class NoSearch:
    def __init__(self, n, d=2, **kwargs):
        self.n = n
        self.d = d

        self.X = np.random.randn(n, d) / 1000
        self.G = self.X @ self.X.T
        # self.G = np.eye(n)
        # self.X = gram_utils.decompose(self.G, d=self.d)

        self.data = []
        self.num_answers = 0
        self.answers = []

    def get_query(self):
        """
        Returns [head, predicted_winner, predicted_loser]
        """
        return random_query(self.n)

    def process_answer(self, head, winner, loser):
        self.num_answers = len(self.answers)
        num_ans = self.num_answers

        start = time()
        self.G = update(self.G, head, winner, loser)

        beta = self.n
        if num_ans > 2 * beta:
            for i in np.random.choice(num_ans, size=min(20, beta - 1),):
                self.G = update(self.G, *self.answers[i])
        self._times = {"update_time": time() - start}
        if len(self.answers) % self.n == 0:
            self.X = gram_utils.decompose(self.G, d=self.d)
        self.answers += [[head, winner, loser]]


class RandomSearch(NoSearch):
    def __init__(self, n, t_max=0.05, R=10, **kwargs):
        super().__init__(n, **kwargs)
        self.t_max = t_max
        self.tau = np.random.rand(n, n)
        self.n = n
        self.R = R
        self._summary = {"t_max": t_max, "name": type(self).__name__}

    def get_query(self):
        if (
            self.num_answers < self.R * self.n
            or self.t_max == 0
            or np.abs(self.t_max) < 0
            or np.allclose(self.t_max, 0)
        ):
            q = random_query(self.n)
            self._saved = {"query": q, "score": -np.inf, "searched": 0}
            return q

        n, tau, G = self.n, self.tau, self.G
        D = gram_utils.distances(G)

        best_score = -np.inf
        start = time()
        searched = 0
        best_q = random_query(n)
        while time() - start < self.t_max:
            searched += 10
            queries = [random_query(n) for _ in range(10)]
            scores = [search.score(q, tau, D) for q in queries]

            best_idx = np.argmax(scores)
            if best_score < scores[best_idx]:
                best_q = queries[best_idx]
                best_score = scores[best_idx]
        self._saved = {"query": best_q, "score": best_score, "searched": searched}
        return best_q

    def process_answer(self, head, winner, loser):
        super().process_answer(head, winner, loser)
        if self.num_answers % 10 == 0:
            D = gram_utils.distances(self.G)
            self.tau = search.posterior(D, self.answers)

        self.data += [
            {
                "num_answers": self.num_answers,
                "dist2[head, winner]": LA.norm(self.X[head] - self.X[winner]) ** 2,
                "dist2[head, loser]": LA.norm(self.X[head] - self.X[loser]) ** 2,
                **self._times,
                **search.decide(self.X, *self._saved["query"])[1],
                **self._saved,
                **self._summary,
            }
        ]



import logging
from time import sleep
from typing import List, Optional, Tuple

import numpy as np

from salmon.backend.sampler import Sampler
from salmon.triplets.samplers.utils import Answer, Query

logger = logging.getLogger(__name__)


def _get_query(targets: List[int]) -> Tuple[int, int, int]:
    a, b, c = np.random.choice(targets, size=3, replace=False)
    return int(a), int(b), int(c)


class Random(Sampler):
    """
    Choose the triplet queries randomly, only ensuring objects are not repeated.

    Parameters
    ----------
    n : int
        Number of objects
    ident : str
        Identifier of the algorithm

    """

    def __init__(self, n, d=2, ident="", targets=None):
        self.n = n
        self.d = d
        self.targets = targets or list(range(n))

        if targets is not None:
            if not isinstance(targets, list):
                msg = "Specify a list for targets. Got {} or type {}"
                raise ValueError(msg.format(targets, type(targets)))
            if not all(isinstance(i, int) for i in targets):
                msg = "Not all items in targets are integers. Bad values are {}"
                bad_vals = [v for v in targets if not isinstance(v, int)]
                raise ValueError(msg.format(bad_vals))
            if len(targets) < 3:
                msg = "Specify at least 3 targets items. Got {} targets items"
                raise ValueError(msg.format(len(targets)))
            if max(targets) >= n:
                msg = "At least one targets target is too large. Values too large include {}, larger than {}"
                bad_vals = [v for v in targets if v >= n]
                raise ValueError(msg.format(bad_vals, n))
        super().__init__(ident=ident)

    def get_query(self, **kwargs) -> Tuple[Query, Optional[float]]:
        h, a, b = _get_query(self.targets)
        query = {"head": int(h), "left": int(a), "right": int(b)}
        return query, -9999

    def process_answers(self, ans: List[Answer]):
        return self, False

    def run(self, *args, **kwargs):
        from rejson import Path
        root = Path.rootPath()

        rj = self.redis_client()
        rj.jsonset(f"stopped-{self.ident}", Path("."), True)
        return None



import itertools

import numpy as np
from sklearn.utils import check_random_state

from salmon.triplets.samplers import TSTE
from salmon.triplets.samplers.adaptive import InfoGainScorer

from .utils import alien_egg


def dataset(n, num_ans=1000, random_state=42):
    rng = check_random_state(random_state)
    X = rng.choice(n, size=(num_ans, 3)).astype(np.int16)
    repeats = (X[:, 0] == X[:, 1]) | (X[:, 1] == X[:, 2]) | (X[:, 0] == X[:, 2])
    X = X[~repeats]
    assert len(X) >= num_ans * 0.9
    y = [alien_egg(*x, random_state=rng) for x in X]
    return X, y


def answer(X, y):
    answers = [
        {"head": h, "left": l, "right": r, "winner": l if yi == 0 else r}
        for (h, l, r), yi in zip(X, y)
    ]
    alg_ans = [
        (a["head"], a["winner"], a["left"] if a["winner"] == a["right"] else a["right"])
        for a in answers
    ]
    return alg_ans


def test_same_salmon_next(n=40, d=2, num_ans=4000):
    X, y = dataset(n, num_ans=num_ans, random_state=42)
    ans = answer(X, y)

    est = TSTE(n=n, d=d)
    new_embedding = (np.arange(n * d) // d).reshape(n, d).astype("float32")

    for s1, s2 in [(30, 10)]:
        new_embedding[s1] = s2
        new_embedding[s2] = s1

    search = InfoGainScorer(embedding=new_embedding, probs=est.opt.module_.probs)
    search.push(ans)

    _, (useful, useless) = search.score(queries=[(29, 30, 10), (29, 11, 28)])
    assert useful > useless

    t = list(range(40))
    queries = [
        (h, o1, o2)
        for h, o1, o2 in itertools.product(t, t, t)
        if h != o1 != o2 and h != o2
    ]
    Q, scores = search.score(queries=queries)
    flipped = [30, 10]
    useful = (
        np.isin(Q[:, 0], flipped)
        | np.isin(Q[:, 1], flipped)
        | np.isin(Q[:, 2], flipped)
    )
    useful = np.isin(Q[:, 1], flipped) & np.isin(Q[:, 2], flipped)
    side_info = [29, 31, 9, 11]
    useful &= np.isin(Q[:, 0], side_info)
    assert scores[useful].mean() > scores[~useful].mean()
    assert np.median(scores[useful]) > np.median(scores[~useful])

    assert np.percentile(scores[~useful], 50) < np.percentile(scores[useful], 23)
    assert np.percentile(scores[~useful], 60) < np.percentile(scores[useful], 25)
    assert np.percentile(scores[~useful], 70) < np.percentile(scores[useful], 30)
    assert np.percentile(scores[~useful], 80) < np.percentile(scores[useful], 50)
    assert np.percentile(scores[~useful], 90) < np.percentile(scores[useful], 75)

    # Another test to make sure that "useless" queries have lower scores than
    # "useful" queries.
    all_queries = set(range(40))
    good_items = sum([list(range(q - 4, q + 4 + 1)) for q in side_info + flipped], [])
    useless = np.isin(Q[:, 0], list(all_queries - set(good_items)))
    useless |= np.isin(Q[:, 1], list(all_queries - set(good_items)))
    useless |= np.isin(Q[:, 2], list(all_queries - set(good_items)))

    assert -1.1 < np.median(scores[useless]) + 0.3 < np.median(scores[useful]) < -0.75
    assert -1.13 < scores[useless].mean() + 0.19 < scores[useful].mean() < -0.93
    assert -1.85 < scores[useless].min() < scores[useful].min() < -1.46

    p = np.percentile
    assert -1.29 < p(scores[useless], 30) < p(scores[useful], 19) < -1.23
    assert -1.17 < p(scores[useless], 40) < p(scores[useful], 23) < 1.04
    assert -1.10 < p(scores[useless], 50) < p(scores[useful], 24) < -0.98
    assert -1.05 < p(scores[useless], 60) < p(scores[useful], 25) < -0.94
    assert -0.93 < p(scores[useless], 70) < p(scores[useful], 26) < -0.90
    assert -0.80 < p(scores[useless], 85) < p(scores[useful], 30) < -0.77
    assert -0.76 < p(scores[useless], 90) < p(scores[useful], 80) < -0.74



import cloudpickle
import numpy as np

from salmon.triplets.samplers import RoundRobin


def test_rr():
    alg = RoundRobin(n=10)
    alg.foo = "bar"
    ir = cloudpickle.dumps(alg)
    alg2 = cloudpickle.loads(ir)
    assert type(alg2) == RoundRobin
    assert alg2.n == 10
    assert alg.foo == alg2.foo
    assert alg.meta_ == alg2.meta_



import itertools
import random
from gc import collect as garbage_collect
from pprint import pprint
from time import sleep, time
from typing import Any, Dict, List, Optional, Tuple, TypeVar

import numpy as np

from salmon.utils import flush_logger, get_logger

logger = get_logger(__name__)

Query = TypeVar("Query")
Answer = TypeVar("Answer")


class Sampler:
    """
    Run a sampling algorithm. Provides hooks to connect with the database and
    the Dask cluster.

    Parameters
    ----------
    ident : str
        The algorithm idenfifier. This value is used to identify the algorithm
        in the database.
    """

    def __init__(self, ident: str = ""):
        self.ident = ident
        self.meta_ = []

    def redis_client(self, decode_responses=True) -> "RedisClient":
        """
        Get the database (/Redis client)
        """
        from rejson import Client as RedisClient

        return RedisClient(host="redis", port=6379, decode_responses=decode_responses)


    def run(self, client: "DaskClient"):
        """
        Run the algorithm.

        Parameters
        ----------
        client : DaskClient
            A client to Dask.
        rj : RedisClient
            A Redist Client, a rejson.Client

        Notes
        -----
        This function runs the adaptive algorithm. Because it's asynchronous,
        this function should return if
        ``"reset" in rj.keys() and rj.jsonget("reset")``.

        """
        import dask.distributed as distributed
        from redis.exceptions import ResponseError
        from rejson import Path
        root = Path.rootPath()

        rj = self.redis_client()

        answers: List = []
        logger.info(f"Staring {self.ident}")

        def submit(fn: str, *args, allow_other_workers=True, **kwargs):
            if "workers" in kwargs:
                kwargs.update({"allow_other_workers": allow_other_workers})
            return client.submit(getattr(type(self), fn), *args, **kwargs,)

        update = False
        queries = np.array([])
        scores = np.array([])
        n_model_updates = 0
        rj.jsonset(f"alg-perf-{self.ident}", root, [])
        save_deadline = 0.0  # right away
        data: List[Dict[str, Any]] = []

        error_raised: List[int] = []
        for k in itertools.count():
            try:
                loop_start = time()
                datum = {"iteration": k, "ident": self.ident, "time": time()}

                answers = self.get_answers(rj, clear=True)
                datum["num_answers"] = len(answers)
                self_future = client.scatter(self)

                _start = time()
                if len(queries) and len(scores):
                    queries_f = client.scatter(queries)
                    scores_f = client.scatter(scores)
                else:
                    queries_f = scores_f = []
                if update:
                    datum["cleared_queries"] = True
                    __start = time()
                    self.clear_queries(rj)
                    datum["time_clearing"] = time() - __start
                else:
                    datum["cleared_queries"] = False
                done = distributed.Event(name="pa_finished")
                done.clear()

                workers = list(client.has_what())
                random.shuffle(workers)
                f_post = submit(
                    "post_queries",
                    self_future,
                    queries_f,
                    scores_f,
                    done=done,
                    workers=workers[0],
                )
                f_model = submit(
                    "process_answers", self_future, answers, workers=workers[1],
                )

                f_search = submit(
                    "get_queries", self_future, stop=done, workers=workers[2],
                )

                time_model = 0.0
                time_post = 0.0
                time_search = 0.0

                def _model_done(_):
                    nonlocal time_model
                    nonlocal done
                    done.set()
                    time_model += time() - _start

                def _post_done(_):
                    nonlocal time_post
                    time_post += time() - _start

                def _search_done(_):
                    nonlocal time_search
                    time_search += time() - _start

                f_model.add_done_callback(_model_done)
                f_post.add_done_callback(_post_done)
                f_search.add_done_callback(_search_done)

                # Future.result raises errors automatically
                posted = f_post.result()
                new_self, update = f_model.result()
                queries, scores, search_meta = f_search.result()

                _datum_update = {
                    "n_queries_posted": posted,
                    "n_queries_scored": len(queries),
                    "n_queries_in_db": rj.zcard(f"alg-{self.ident}-queries"),
                    "model_updated": update,
                    "n_model_updates": n_model_updates,
                    "time_posting_queries": time_post,
                    "time_model_update": time_model,
                    "time_search": time_search,
                    "time": time(),
                    **search_meta,
                }
                datum.update(_datum_update)
                if update:
                    _s = time()
                    self.__dict__.update(new_self.__dict__)
                    datum["time_update"] = time() - _s
                    n_model_updates += 1

                if time() > save_deadline + 1e-3:
                    save_deadline = time() + 60
                    _s = time()
                    self.save()
                    datum["time_save"] = time() - _s
                datum["time_loop"] = time() - loop_start

                data.append(datum)
                logger.info(datum)
                posting_deadline = data[0]["time"] + 2 * 60
                if time() >= posting_deadline or k == 10 or k == 20:
                    flush_logger(logger)
                    keys = data[-1].keys()
                    to_post = {}
                    for _k in keys:
                        vals = [d.get(_k, None) for d in data]
                        vals = [v for v in vals if v]
                        if not len(vals):
                            continue
                        if isinstance(vals[0], (int, np.integer)):
                            Type = int
                        elif isinstance(vals[0], (float, np.floating)):
                            Type = float
                        else:
                            continue
                        _update = {
                            f"{_k}_median": np.median(vals),
                            f"{_k}_mean": np.mean(vals),
                            f"{_k}_min": np.min(vals),
                            f"{_k}_max": np.max(vals),
                        }
                        if _k == "time":
                            _update = {"time": _update["time_median"]}
                        to_post.update({_k: Type(v) for _k, v in _update.items()})

                    try:
                        rj.jsonarrappend(f"alg-perf-{self.ident}", root, to_post)
                    except ResponseError as e:
                        if (
                            "could not perform this operation on a key that doesn't exist"
                            in str(e)
                        ):
                            # I think this happens when the frontend deletes
                            # the database when /reset is triggered
                            pass
                        else:
                            raise e

                    data = []

                if "reset" in rj.keys() and rj.jsonget("reset", root):
                    logger.warning(f"Resetting {self.ident}")
                    self.reset(client, rj, futures=[f_model, f_post, f_search])
                    break

            except Exception as e:
                logger.exception(e)
                flush_logger(logger)
                error_raised.append(k)

                __n = 5
                if np.diff(error_raised[-__n:]).tolist() == [1] * (__n - 1):
                    logger.exception(e)
                    flush_logger(logger)
                    raise e
        return True

    def save(self) -> bool:
        """
        Save the sampler's state and current embedding to the database.
        """
        import cloudpickle

        rj2 = self.redis_client(decode_responses=False)

        out = cloudpickle.dumps(self)
        rj2.set(f"state-{self.ident}", out)

        try:
            out = cloudpickle.dumps(self.get_model())
        except NotImplementedError:
            pass
        else:
            rj2.set(f"model-{self.ident}", out)
        return True

    def reset(self, client, rj, futures=None):
        """
        Stop the algorithm. The algorithm will be deleted shortly after
        this function is called.
        """
        from rejson import Client as RedisClient
        from rejson import Path
        root = Path.rootPath()

        reset = rj.jsonget("reset", root)
        logger.warning("reset=%s for %s", reset, self.ident)
        if not reset:
            return False

        logger.warning(f"Deleting various keys for {self.ident}")
        rj2 = RedisClient(host="redis", port=6379, decode_responses=False)
        rj2.delete(f"state-{self.ident}")
        rj2.delete(f"model-{self.ident}")
        rj.jsondel(f"alg-perf-{self.ident}", root)
        rj.delete(f"alg-perf-{self.ident}")

        # Clear answers
        logger.warning(f"Clearing answers for {self.ident}")
        self.get_answers(rj, clear=True)

        # Clear queries (twice)
        logger.warning(f"Clearing queries for {self.ident}")
        key = f"alg-{self.ident}-queries"
        for k in range(4, 18):
            limit = 2 ** k
            rj.zremrangebyscore(key, -limit, limit)
            sleep(0.1)
            n_queries = rj.zcard(key)
            logger.warning(f"n_queries={n_queries}")
            if not n_queries:
                break
        logger.warning(f"Clearing queries again for {self.ident}")
        self.clear_queries(rj)

        if futures:
            for future in futures:
                if future:
                    client.cancel(future, force=True)

        logger.warning(f"Restarting Dask client for {self.ident}")
        f = client.restart(timeout="5s")
        try:
            client.sync(f)
        except:
            pass

        client.run(garbage_collect)

        logger.warning(f"Setting stopped-{self.ident}")
        rj.jsonset(f"stopped-{self.ident}", Path("."), True)
        logger.warning(f"All done stopping {self.ident}")
        return True

    @property
    def clear(self):
        """
        Should the queries be cleared from the database?
        """
        return True

    def process_answers(self, answers: List[Answer]):
        """
        Process answers.

        Parameters
        ----------
        answers : List[Answers]
            Each answer is a dictionary. Each answer certainly has the keys
            "head", "left", "right" and "winner", and may have the key
            "puid" for participant UID.

        Returns
        -------
        data : dict
            An update to self.__dict__.
        """
        raise NotImplementedError

        def get_queries(self) -> Tuple[List[Query], List[float]]:
            """
            Get queries.

            Returns
            -------
            queries : List[Query]
                The list of queries
            scores : List[float]
                The scores for each query. Higher scores are sampled more
                often.
            meta : Dict[str, Any]
                Information about the search.

            Notes
            -----
            The scores have to be unique. The underlying implementation does
            not sample queries of the same score unbiased.

            """
            return [], [], {}

    def get_model(self) -> Dict[str, Any]:
        """
        Get the model underlying the algorithm.

        Returns
        -------
        state : Dict[str, Any]
            The state of the algorithm. This can be used for display on the
            dashboard or with an HTTP get request.
        """
        raise NotImplementedError

    def clear_queries(self, rj: "RedisClient") -> bool:
        """
        Clear all queries that this sampler has posted from the database.
        """
        rj.delete(f"alg-{self.ident}-queries")
        return True

    def post_queries(
        self,
        queries: List[Query],
        scores: List[float],
        rj: Optional["RedisClient"] = None,
        done=None,
    ) -> int:
        """
        Post scored queries to the database.

        Parameters
        ----------
        queries : List[Query]
            Queries to post to the database
        scores : List[float]
            The scores for each query
        rj : RedisClient, optional
            The databaase

        Returns
        -------
        n_queries : int
            The number of queries posted to the database.
        """
        if rj is None:
            rj = self.redis_client()

        if not len(queries):
            return 0

        if isinstance(queries, np.ndarray) and isinstance(scores, np.ndarray):
            idx = np.argsort(-1 * scores)
            assert (
                len(scores) == queries.shape[0]
            ), f"Different lengths {scores.shape}, {queries.shape}"

            scores = scores[idx]  # high to low scores
            queries = queries[idx]
            valid = ~np.isnan(scores)
            scores = scores[valid]
            queries = queries[valid]
            high = scores[0]
            low = scores[-1]
            assert low <= high, f"high={high} to low={low} scores"

        chunk_size = 2000
        n_chunks = len(queries) // chunk_size
        split_queries = np.array_split(queries, max(n_chunks, 1))
        split_scores = np.array_split(scores, max(n_chunks, 1))

        n_queries = 0
        name = self.ident
        key = f"alg-{name}-queries"
        for _queries, _scores in zip(split_queries, split_scores):
            queries2 = {
                self.serialize_query(q): float(s)
                for q, s in zip(_queries, _scores)
                if not np.isnan(s)
            }
            if len(queries2):
                rj.zadd(key, queries2)
            n_queries += len(queries2)
            if done is not None and done.is_set():
                break

        return n_queries

    def serialize_query(self, q: Query) -> str:
        """
        Serialize a query (so it can go in the database).
        """
        # TODO: use ast.literal_eval or json.loads
        h, a, b = q
        return f"{h}-{a}-{b}"

    def get_answers(self, rj: "RedisClient", clear: bool = True) -> List[Answer]:
        """
        Get all answers the frontend has received.
        """
        from rejson import Path
        root = Path.rootPath()

        if not clear:
            raise NotImplementedError
        key = f"alg-{self.ident}-answers"
        if key in rj.keys():
            pipe = rj.pipeline()
            pipe.jsonget(key, Path("."))
            pipe.jsonset(key, Path("."), [])
            answers, success = pipe.execute()
            return answers
        return []



import numpy as np
from sklearn.base import BaseEstimator

import salmon.utils as utils
from salmon.triplets.samplers.adaptive.search import gram_utils, score

logger = utils.get_logger(__name__)


class QueryScorer:
    """
    A class to score queries for adaptive searches.

    Parameters
    ----------
    embedding : array-like
        The embedding of points.

    probs : callable
        Function to call to get probabilities. Called with
        ``probs(win2, lose2)``
        where ``win2`` and ``lose2`` are the squared Euclidean distances
        between the winner and loser.

    Notes
    -----
    Inputs: include an embedding, noise model and history of answers
    received.

    Outputs: query scores.

    Internal state: the posterior, history.

    Public API:

    * Update posterior.
    * Get scores.

    """

    def __init__(self, embedding=None, probs=None):
        self.embedding = embedding
        self.probs = probs

    def _initialize(self):
        self.initialized_ = True
        n = len(self.embedding)
        self._tau_ = np.zeros((n, n), dtype="float32")
        self.push([])

    def _random_queries(self, n, num=1000, trim=True):
        new_num = int(num * 1.1 + 3)
        rng = np.random.RandomState()
        queries = rng.choice(n, size=(new_num, 3))
        repeated = (
            (queries[:, 0] == queries[:, 1])
            | (queries[:, 1] == queries[:, 2])
            | (queries[:, 0] == queries[:, 2])
        )
        queries = queries[~repeated]
        if trim:
            return queries[:num]
        return queries

    def _distances(self):
        G = gram_utils.gram_matrix(self.embedding)
        return gram_utils.distances(G)

    def score(self):
        raise NotImplementedError

    def _posterior(self, D, history):
        """
        Calculate the posterior.

        Parameters
        ----------
        D : array-like
            Distance array. D[i, j] = ||x_i - x_j||_2^2
        S : array-like, shape=(num_ans, 3)
            History of answers.

        Returns
        -------
        posterior : array-like, shape=(self.n, self.n)
        """
        n = D.shape[0]

        logger.info("history = %s", history)
        if len(history):
            for k, (head, w, l) in enumerate(history):
                probs = self.probs(D[w], D[l])
                probs[np.isnan(probs)] = 0
                _eps = 1e-80
                probs[probs <= _eps] = _eps
                a = np.log(probs + _eps)
                self._tau_[head] += a

        # idx = self._tau_ >= -np.inf
        tau = np.zeros_like(self._tau_)
        # tau[idx] = np.exp(self._tau_[idx])
        tau = np.exp(self._tau_)
        s = tau.sum(axis=1)  # the sum of each row

        gt0 = s > 0
        eps = s[gt0].min() if gt0.any() else s.min()

        s *= 1e4
        tau *= 1e4

        s = np.clip(s, eps, np.inf)

        tau2 = (tau.T / s).T  # transpose to make broadcasting work
        return tau2

    def push(self, history):
        if not hasattr(self, "initialized_"):
            self._initialize()
        D = self._distances()
        self.posterior_ = self._posterior(D, history)
        return self


class InfoGainScorer(QueryScorer):
    def score(self, *, queries=None, num=1000):
        """
        Score the queries using (almost) the information gain.

        Parameters
        ----------
        queries : List[int, int, int]
            The list of queries to score.
        num : int
            Number of random queries to generate.

        """
        if not hasattr(self, "initialized_"):
            self._initialize()
        D = self._distances()
        if queries is not None and num != 1000:
            raise ValueError("Only specify one of `queries` or `num`")
        if queries is None:
            queries = self._random_queries(len(self.embedding), num=num)
        Q = np.array(queries).astype("int64")
        H, O1, O2 = Q[:, 0], Q[:, 1], Q[:, 2]

        scores = score(H, O1, O2, self.posterior_, D, probs=self.probs)
        return Q, scores


class UncertaintyScorer(QueryScorer):
    def score(self, *, queries=None, num=1000, trim=True):
        """
        Score the queries using (almost) the information gain.

        Parameters
        ----------
        queries : List[int, int, int]
            The list of queries to score.
        num : int
            Number of random queries to generate.

        """
        if not hasattr(self, "initialized_"):
            self._initialize()
        D = self._distances()
        if queries is not None and num != 1000:
            raise ValueError("Only specify one of `queries` or `num`")
        if queries is None:
            queries = self._random_queries(len(self.embedding), num=num, trim=trim)
        Q = np.array(queries).astype("int64")
        H, O1, O2 = Q[:, 0], Q[:, 1], Q[:, 2]

        # score is distance to the decision boundary.
        scores = -np.abs(D[H, O1] - D[H, O2])
        return Q, scores



import ast
import json
import os
import pickle
import random
from datetime import datetime, timedelta
from pathlib import Path
from time import sleep, time
from typing import Tuple

import numpy as np
import pandas as pd
import pytest
import yaml

from salmon.triplets.manager import Config

from .utils import LogError, logs, server


def test_basics(server, logs):
    """
    Requires `docker-compose up` in salmon directory
    """
    server._authorized = False  # mock not unauthorized
    server.delete("/reset", status_code=401, timeout=20)
    server.authorize()
    r = server.delete("/reset?force=1", timeout=20)
    assert r.json() == {"success": True}
    server.delete("/reset", status_code=401, auth=("foo", "bar"))

    r = server.authorize()
    assert r.status_code == 200
    r = server.delete("/reset?force=1", timeout=20)
    server.authorize()

    assert r.json() == {"success": True}
    server.get("/init")
    exp = Path(__file__).parent / "data" / "exp.yaml"
    server.post("/init_exp", data={"exp": exp.read_text()})
    exp_config = yaml.safe_load(exp.read_bytes())
    puid = "puid-foo"
    answers = []
    print("Starting loop...")
    n_ans = 40
    for k in range(n_ans):
        _start = time()
        q = server.get("/query").json()
        ans = {"winner": random.choice([q["left"], q["right"]]), "puid": puid, **q}
        answers.append(ans)
        sleep(10e-3)
        ans["response_time"] = time() - _start
        server.post("/answer", data=ans)

    r = server.get("/responses")
    for server_ans, actual_ans in zip(r.json(), answers):
        assert set(actual_ans).issubset(server_ans)
        assert all(
            actual_ans[k] == server_ans[k] for k in ["head", "winner", "left", "right"]
        )
        assert server_ans["puid"] == puid
    df = pd.DataFrame(r.json())
    expected_cols = {
        "start_time",
        "time_received_since_start",
        "time_received",
        "head",
        "right",
        "left",
        "winner",
        "loser",
        "winner_html",
        "loser_html",
        "right_html",
        "head_html",
        "left_html",
        "winner_filename",
        "right_filename",
        "left_filename",
        "loser_filename",
        "head_filename",
        "puid",
        "response_time",
        "network_latency",
        "datetime_received",
        "sampler",
        "score",
        "puid_num_responses",
        "num_responses",
    }
    assert (df["winner"] != df["loser"]).all()
    assert ((df["winner"] == df["left"]) | (df["winner"] == df["right"])).all()
    assert ((df["loser"] == df["left"]) | (df["loser"] == df["right"])).all()
    n = len(exp_config["targets"])
    assert (0 == df["head"].min()) and (df["head"].max() == n - 1)
    assert ((0 == df["left"].min()) or (df["right"].min() == 0)) and (
        (df["left"].max() == n - 1) or (df["right"].max() == n - 1)
    )
    assert 10e-3 < df.response_time.min()
    assert expected_cols == set(df.columns)
    assert df.puid.nunique() == 1

    assert np.allclose(df.score, -9999)  # -9999 is a proxy for nan here

    # Make sure ordered by time
    assert df.time_received_since_start.diff().min() > 0

    r = server.get("/responses")
    assert r.status_code == 200
    assert "exception" not in r.text
    df = pd.DataFrame(r.json())
    assert len(df) == n_ans

    r = server.get("/dashboard")
    assert r.status_code == 200
    assert "Embedding dimension: 2" in r.text
    assert "Number of targets: 6" in r.text
    assert "Samplers: [&#39;random&#39;]" in r.text  # &#39; is HTML for the apostrophe

    assert (df["num_responses"] == (np.arange(len(df)) + 1)).all()


def test_bad_file_upload(server):
    server.authorize()
    server.get("/init")
    exp = Path(__file__).parent / "data" / "bad_exp.yaml"
    r = server.post("/init_exp", data={"exp": exp.read_text()}, error=True)
    assert r.status_code == 500
    assert "Error" in r.text
    assert "yaml" in r.text
    assert "-\tfoo" in r.text


def test_no_repeats(server):
    server.authorize()
    exp = Path(__file__).parent / "data" / "exp.yaml"
    server.post("/init_exp", data={"exp": exp.read_text()})
    for k in range(50):
        q = server.get("/query").json()
        ans = {"winner": random.choice([q["left"], q["right"]]), "puid": "foo", **q}
        server.post("/answer", data=ans)
        sleep(10e-3)

    r = server.get("/responses")
    df = pd.DataFrame(r.json())
    equal_targets = (
        (df["head"] == df["right"])
        | (df["head"] == df["left"])
        | (df["left"] == df["right"])
    )
    assert not equal_targets.all()


def test_meta(server):
    server.authorize()
    exp = Path(__file__).parent / "data" / "exp.yaml"
    server.post("/init_exp", data={"exp": exp.read_text()})
    num_ans = 10
    for k in range(num_ans):
        q = server.get("/query").json()
        ans = {"winner": random.choice([q["left"], q["right"]]), "puid": str(k), **q}
        server.post("/answer", data=ans)
    meta = server.get("/meta").json()
    assert meta["participants"] == num_ans
    assert meta["responses"] == num_ans


def test_saves_state(server):
    server.authorize()
    server.delete("/reset?force=1", timeout=20)
    sleep(0.1)

    this_dir = Path(__file__).absolute().parent
    root = this_dir.parent
    dump_dir = root / "salmon" / "_out"
    assert dump_dir.exists()
    dump = dump_dir / "dump.rdb"
    assert not dump.exists()
    exp = this_dir / "data" / "exp.yaml"

    server.authorize()
    server.post("/init_exp", data={"exp": exp.read_text()})
    for k in range(10):
        q = server.get("/query").json()
        ans = {"winner": random.choice([q["left"], q["right"]]), "puid": str(k), **q}
        server.post("/answer", data=ans)
    r = server.get("/dashboard")
    assert r.status_code == 200
    sleep(1)
    assert dump.exists()

    # Clear all dump files; reset state
    dump_files = list(dump_dir.glob("*.rdb"))
    for d in dump_files:
        d.unlink()
    files = [f.name for f in dump_dir.glob("*.rdb")]
    assert len(files) == 0

    # Make sure saved resetting saves experiment state
    before_reset = datetime.now()
    server.delete("/reset?force=1", timeout=20)
    files = [f.name for f in dump_dir.glob("*.rdb")]
    assert len(files) == 1
    written = datetime.strptime(files[0], "dump-%Y-%m-%dT%H:%M.rdb")
    assert isinstance(written, datetime)

    # Because docker container time zones are screwy... this is a check to
    # make sure this new file was written sometime today
    day = timedelta(hours=24)
    assert datetime.now() - day < written < datetime.now() + day


def test_download_restore(server):
    server.authorize()
    dump = Path(__file__).absolute().parent.parent / "salmon" / "_out" / "dump.rdb"
    assert not dump.exists()
    exp = Path(__file__).parent / "data" / "exp.yaml"
    server.post("/init_exp", data={"exp": exp.read_text()})
    data = []
    for k in range(10):
        q = server.get("/query").json()
        ans = {"winner": random.choice([q["left"], q["right"]]), "puid": str(k), **q}
        server.post("/answer", data=ans)
        data.append(ans)
    r = server.get("/download")
    assert all(x in r.headers["content-disposition"] for x in ["exp-", ".rdb"])

    # Does it restore?
    content = dump.read_bytes()
    dump.unlink()
    assert not dump.exists()
    server.post("/restore", data={"rdb": content})
    assert dump.exists()


def test_logs(server, logs):
    server.authorize()
    dump = Path(__file__).absolute().parent.parent / "salmon" / "out" / "dump.rdb"
    assert not dump.exists()
    exp = Path(__file__).parent / "data" / "exp.yaml"
    server.delete("/reset?force=1", timeout=20)
    server.authorize()
    server.post("/init_exp", data={"exp": exp.read_text()})
    data = []
    puid = "adsfjkl4awjklra"
    for k in range(10):
        q = server.get("/query").json()
        ans = {"winner": random.choice([q["left"], q["right"]]), "puid": k, **q}
        server.post("/answer", data=ans)
        data.append(ans)

    r = server.get("/logs")
    assert r.status_code == 200
    logs = r.json()
    query_logs = logs["salmon.frontend.public.log"]

    str_answers = [q.strip("\n") for q in query_logs if "answer received" in q]
    answers = [ast.literal_eval(q[q.find("{") :]) for q in str_answers]
    puids = {ans["puid"] for ans in answers}
    assert {str(i) for i in range(10)}.issubset(puids)


def test_zip_upload(server):
    server.authorize()
    exp = Path(__file__).parent / "data" / "upload_w_zip.yaml"
    targets = Path(__file__).parent / "data" / "imgs.zip"
    assert targets.exists()
    t = targets.read_bytes()
    assert len(t) > 0
    assert t[:4] == b"\x50\x4B\x03\x04"
    server.post("/init_exp", data={"exp": exp.read_text()}, files={"targets": t})


def test_get_config(server):
    server.authorize()
    exp = Path(__file__).parent / "data" / "exp.yaml"
    server.post("/init_exp", data={"exp": exp.read_text()})
    my_config = yaml.safe_load(exp.read_text())
    rendered_config = server.get("/config").json()
    assert set(my_config).issubset(rendered_config)
    user_html = my_config.pop("html")
    assert set(user_html).issubset(set(rendered_config["html"]))
    for k, v in my_config.items():
        if k == "targets":
            v = [str(t) for t in v]
        assert rendered_config[k] == v

    yaml_config = server.get("/config?json=0").text
    assert "\n" in yaml_config
    assert yaml.safe_load(yaml_config) == rendered_config


def test_config_defaults_update(server):
    server.authorize()
    html = {
        "instructions": "foo",
        "debrief": "bar",
        "max_queries": 42,
        "custom_tag": "foo",
    }
    exp = {"targets": 10, "html": html}
    server.post("/init_exp", data={"exp": exp})

    rendered = server.get("/config").json()
    assert "custom_tag" in rendered["html"] and rendered["html"]["custom_tag"] == "foo"
    assert rendered["html"]["instructions"] == "foo"
    assert rendered["html"]["debrief"] == "bar"
    assert rendered["html"]["max_queries"] == 42
    assert rendered["samplers"] == {"random": {"class": "Random"}}


def test_config_misplaced_error(server):
    server.authorize()
    exp = {"instructions": "foo", "debrief": "bar", "max_queries": 42, "targets": 10}
    r = server.post("/init_exp", data={"exp": exp}, error=500)
    assert "Move keys" in r.text and "YAML" in r.text
    assert "include this block of YAML" in r.text and "html:\n  debrief: bar" in r.text


def test_no_init_twice(server, logs):
    """
    Requires `docker-compose up` in salmon directory
    """
    server.authorize()
    exp = Path(__file__).parent / "data" / "exp.yaml"
    server.post("/init_exp", data={"exp": exp.read_text()}, timeout=20)
    query = server.get("/query")
    assert query

    # Make sure errors on re-initialization
    server.post("/init_exp", data={"exp": exp.read_text()}, status_code=403, timeout=20)

    # Make sure the prescribed method works (resetting, then re-init'ing)
    server.delete("/reset", status_code=403, timeout=20)
    server.delete("/reset?force=1", timeout=20)

    server.authorize()
    server.post("/init_exp", data={"exp": exp.read_text()}, timeout=20)
    query = server.get("/query")
    assert query


def test_auth_repeated_entries(server):
    server._authorized = False
    server.post("/init_exp", status_code=401)  # unauthorized
    name, pword = "dfjklasdfsdf32", "baz"
    server._authorized = True
    r = server.post(f"/create_user/{name}/{pword}", status_code=403)
    assert "maximum number of users" in r.text.lower()


def test_random_error(server, logs):
    n_val = 5
    exp = {"targets": [0, 1, 2, 3, 4, 5], "samplers": {"RandomSampling": {}, "ARR": {}}}

    with pytest.raises(LogError), logs:
        server.authorize()
        r = server.post("/init_exp", data={"exp": exp}, status_code=500)
        assert "The sampler `RandomSampling` has been renamed to `Random`" in r.text


def test_html_defaults_rendered(server):
    server.authorize()
    exp = {
        "targets": 10,
        "html": {"arrow_keys": True, "css": "/*css foo*/", "skip_button": True},
    }
    server.post("/init_exp", data={"exp": exp})
    rendered = server.get("/").text

    html = Config().html.dict()

    assert f"<title>{html['title']}</title>" in rendered
    assert f"<p id=\"instructions\">{html['instructions']}</p>" in rendered
    assert f"id=\"debrief\">{html['debrief']}" in rendered
    assert f"var max_queries = {html['max_queries']};" in rendered
    assert 'id="skip-button"' in rendered and "Skip this question" in rendered
    assert "/*css foo*/" in rendered
    assert (
        "document.onkeydown = function checkKey(e) {" in rendered
        and "e.keyCode == '37') { // left arrow" in rendered
    )


def test_defaults_acceptable_config(server):
    server.authorize()
    r = server.post("/init_exp", data={"exp": {"targets": 10}})
    assert r.status_code == 200
    config = server.get("/config").json()

    server.reset()
    server.authorize()
    r = server.post("/init_exp", data={"exp": config})
    assert r.status_code == 200
    config2 = server.get("/config").json()
    assert config == config2



import asyncio
import json
import os
import pickle
import random
from datetime import datetime, timedelta
from pathlib import Path
from time import sleep, time
from typing import Tuple

import numpy as np
import pandas as pd
import pytest
import yaml

from .utils import LogError, logs, server


def test_samplers_per_user(server, logs):
    exp = Path(__file__).parent / "data" / "active.yaml"
    print("init'ing exp")
    exp2 = yaml.safe_load(exp.read_bytes())
    exp2["sampling"] = {"samplers_per_user": 1}

    print("done")

    with open(exp, "r") as f:
        config = yaml.load(f, Loader=yaml.SafeLoader)
    samplers = list(config["samplers"].keys())

    sampler = random.choice(samplers)
    with logs:
        server.authorize()
        server.post("/init_exp", data={"exp": str(exp2)})
        for k in range(len(samplers) * 2):
            q = server.get(f"/query?sampler={sampler}").json()
            ans = {"winner": random.choice([q["left"], q["right"]]), "puid": "foo", **q}
            server.post("/answer", json=ans)
            sleep(0.1)

        r = server.get("/responses")
        d = r.json()
        df = pd.DataFrame(d)
        algs = df.sampler.unique()
        assert len(set(algs)) == 1
        _ = len(set(algs))


def test_active_wrong_proportion(server, logs):
    exp = {
        "targets": 16,
        "sampling": {"probs": {"a1": 50, "a2": 40}},
        "samplers": {"a1": {"class": "Random"}, "a2": {"class": "Random"},},
    }
    with pytest.raises(LogError):
        with logs:
            server.authorize()
            r = server.post("/init_exp", data={"exp": json.dumps(exp)}, error=True)
            assert r.status_code == 500
            assert "values in sampling.probs should add up to 100" in r.text


def test_active_bad_keys(server, logs):
    exp = {
        "targets": 16,
        "sampling": {"probs": {"a1": 50, "a2": 40}},
        "samplers": {"a1": {"class": "Random"}},
    }
    with pytest.raises(LogError):
        with logs:
            server.authorize()
            r = server.post("/init_exp", data={"exp": exp}, error=True)
            assert r.status_code == 500
            assert all(
                x in r.text.lower()
                for x in ["sampling.probs keys", "are not the same as samplers keys"]
            )


@pytest.mark.parametrize("sampler", ["ARR", "Random"])
def test_active_queries_generated(server, sampler, logs):
    # R=1 chosen because that determines when active sampling starts; this
    # test is designed to make sure no unexpected errors are thrown in
    # active portion (not that it generates a good embedding)

    # tests ARR to make sure active scores are generated;
    # tests Random to make sure that's not a false positive and
    # random queries are properly identifies

    n = 6
    config = {
        "targets": [_ for _ in range(n)],
        "samplers": {sampler: {}},
        "sampling": {},
    }
    if sampler != "Random":
        config["sampling"]["common"] = {"d": 1, "R": 1}
    with logs:
        server.authorize()
        server.post("/init_exp", data={"exp": config})
        active_queries_generated = False
        for k in range(10 * n + 1):
            q = server.get("/query").json()
            query = "random" if q["score"] == -9999 else "active"
            if query == "active":
                active_queries_generated = True
                break

            sleep(200e-3)

            ans = random.choice([q["left"], q["right"]])
            ans = {"winner": ans, "puid": "foo", **q}
            server.post("/answer", json=ans)

            if k % n == 0:
                sleep(1)

    if sampler == "Random":
        assert not active_queries_generated
    else:
        assert active_queries_generated


def test_active_basics(server, logs):
    exp = Path(__file__).parent / "data" / "active.yaml"
    print("init'ing exp")
    print("done")

    with open(exp, "r") as f:
        config = yaml.load(f, Loader=yaml.SafeLoader)
    samplers = list(config["samplers"].keys())

    with logs:
        server.authorize()
        server.post("/init_exp", data={"exp": exp.read_text()})
        for k in range(len(samplers) * 3):
            print(k)
            q = server.get("/query").json()

            # Jerry rig this so this test isn't random (an algorithm is chosen at random)
            q["sampler"] = samplers[k % len(samplers)]
            ans = random.choice([q["left"], q["right"]])

            ans = {"winner": ans, "puid": "foo", **q}
            server.post("/answer", json=ans)

        r = server.get("/responses")
        d = r.json()
        df = pd.DataFrame(d)
        assert (df["score"] <= 1).all()
        algs = df.sampler.unique()
        assert set(algs) == {"TSTE", "ARR", "CKL", "tste2", "GNMDS"}
        assert True  # to see if a log error is caught in the traceback



from typing import Union

import numpy as np
import torch
import torch.nn as nn

ArrayLike = Union[np.ndarray, torch.Tensor]


class TripletDist(nn.Module):
    """
    A base class to find losses for the triplet embedding problem.

    Parameters
    ----------
    n : int
    d : int

    Attributes
    ----------
    embedding : np.ndarray
        The current embedding.

    """

    def __init__(self, n: int = None, d: int = 2, random_state=None):
        super().__init__()
        self.n = n
        self.d = d
        rng = np.random.RandomState(seed=random_state)
        embedding = 1e-4 * rng.randn(n, d).astype("float32")
        self._embedding = torch.nn.Parameter(
            torch.from_numpy(embedding), requires_grad=True
        )

    def numpy_or_torch(self, f):
        def wrapper(win2, lose2):
            converted = False
            if isinstance(win2, np.ndarray):
                win2 = torch.from_numpy(win2)
                lose2 = torch.from_numpy(lose2)
                with torch.no_grad():
                    ret = f(win2, lose2)
                    return ret.detach().numpy()
            return f(win2, lose2)

        return wrapper

    def losses(self, win2: ArrayLike, lose2: ArrayLike) -> ArrayLike:
        """
        Calculate the losses of a this triplet model with the triplet
        distances ``win2`` and ``lose2``. By default, the negative log
        loss of ``self.probs(win2, lose2)``.

        Parameters
        ----------
        win2 : torch.Tensor, shape=(num_answers)
            The squared Euclidean distance between the head vector and
            winner vector. Formally, :math:`\|x_h - x_w\|_2^2`.
        lose2 : torch.Tensor, shape=(num_answers)
            The squared Euclidean distance between the head vector and
            loser vector. Formally, :math:`\|x_h - x_l\|_2^2`.

        Returns
        -------
        prob : torch.Tensor, shape=(num_answers)
            The probability of triplets with those distances being satisfied.
        """
        return -1 * torch.log(self.probs(win2, lose2))

    @property
    def embedding(self):
        return self._embedding

    def probs(self, win2, lose2):
        return self.numpy_or_torch(self._probs)(win2, lose2)

    def _get_dists(self, h_w_l):
        H_W_L = h_w_l.T
        h, w, l = H_W_L[0], H_W_L[1], H_W_L[2]
        heads = self._embedding[h]
        winners = self._embedding[w]
        losers = self._embedding[l]

        win_dist2 = torch.norm(heads - winners, dim=1) ** 2
        lose_dist2 = torch.norm(heads - losers, dim=1) ** 2
        return win_dist2, lose_dist2

    def forward(self, h_w_l: ArrayLike, y=None, sample_weight=None) -> ArrayLike:
        """
        Calculate the probability of a triplet being satisified

        Parameters
        ----------
        h_w_l : torch.Tensor, shape=(num_answers, 3)
            Each row in this 2D array is (head, winner, loser) from
            triplet query.
        y : None, ignored.
        sample_weight : None, ignored.

        Returns
        -------
        losses : torch.Tensor, shape=(num_answers)
            The loss of each individual triplet.
        """
        win2, lose2 = self._get_dists(h_w_l)
        return self.losses(win2, lose2)


class STE(TripletDist):
    def _probs(self, win2, lose2):
        ## Double the computation
        #  num = torch.exp(-win2)
        #  return num / (num + torch.exp(-lose2))

        ## Less computation
        # dist>0: agrees with embedding. <0: does not agree.
        # d1 = win2, d2 = lose2
        # 1 / (1 + exp(d1 - d2))
        # = 1 / (1 + exp(-d2 + d1))
        # = 1 / (1 + exp(-d2 / -d1))
        # = exp(-d1) / (exp(-d1) + exp(d2))
        # = prob of winning by STE
        return 1 / (1 + torch.exp(win2 - lose2))


class TSTE(TripletDist):
    """
    For details
    """

    def __init__(self, n=None, d=2, alpha=1, random_state=None):
        super().__init__(n=n, d=d, random_state=random_state)
        self.alpha = alpha

    def _probs(self, win2, lose2):
        pwr = -(self.alpha + 1) / 2
        t1 = (1 + (win2 / self.alpha)) ** pwr
        t2 = (1 + (lose2 / self.alpha)) ** pwr
        return t1 / (t1 + t2)


class CKL(TripletDist):
    """
    The crowd kernel embedding.
    """

    def __init__(self, n=None, d=2, mu=0.05, random_state=None):
        super().__init__(n=n, d=d, random_state=random_state)
        self.mu = mu

    def _probs(self, win2, lose2):
        num = self.mu + lose2
        return num / (num + self.mu + win2)


class GNMDS(TripletDist):
    """
    The global non-metric multidimensional scaling algorithm.
    """

    def losses(self, win2, lose2):
        zeros = torch.zeros(len(win2))
        return torch.max(zeros, win2 - lose2 + 1)

    def _probs(self, win2, lose2):
        return win2 / (win2 + lose2 + 1e-6)


class SOE(GNMDS):
    def losses(self, win2, lose2):
        zeros = torch.zeros(len(win2))
        return torch.max(zeros, torch.sqrt(win2) - torch.sqrt(lose2) + 1)


class Logistic(GNMDS):
    r"""
    .. math::

       \text{loss}(x, y) = \sum_i \frac{\log(1 + \exp(-y[i]*x[i]))}{\text{x.nelement}()}
    """

    def losses(self, win2: torch.Tensor, lose2: torch.Tensor) -> torch.Tensor:
        # low loss if agrees
        # high loss if disagrees
        # win2 - lose2: positive if disagrees, negative if agrees
        # embedding accurate -> win2 < lose2 -> win2-lose2 < 0 => negative
        # embedding bad -> win2 > lose2 -> win2-lose2 > 0 => positive
        # exp(x): large if x negative, small if x negative.
        _pwrs = torch.cat((torch.tensor([0]), win2 - lose2))
        loss = torch.logsumexp(_pwrs)
        return loss



import random
from pathlib import Path
from time import sleep

import pandas as pd
import pytest
import yaml

from .utils import LogError, logs, server


def test_init_w_queries_valid_idx(server, logs):
    samplers = {"Validation": {"queries": [[0, 1, 4], [0, 1, 5]]}}
    exp = {"targets": [0, 1, 2, 3], "samplers": samplers}

    with pytest.raises(LogError), logs:
        server.authorize()
        server.get("/init")
        r = server.post("/init_exp", data={"exp": exp}, error=True)
        assert r.status_code == 500
        assert "the index 5 is included" in r.text.lower()
        assert "the n=4 targets" in r.text.lower()


def test_validation_sampling(server, logs):
    n_val = 4
    exp = {
        "targets": list(range(10)),
        "samplers": {"Validation": {"n_queries": n_val}},
    }
    data = []
    puid = "adsfjkl4awjklra"

    n_repeat = 4
    server.authorize()
    server.post("/init_exp", data={"exp": exp})
    Q = []
    for k in range(n_repeat * n_val):
        q = server.get("/query").json()
        _ans = random.choice([q["left"], q["right"]])
        ans = {"winner": _ans, "puid": k, **q}
        print([ans[k] for k in ["head", "left", "right"]])
        Q.append(ans)
        server.post("/answer", data=ans)
        data.append(ans)
        sleep(0.20)

    # Test the number of unique queries is specified by n_val
    queries = [(q["head"], (q["left"], q["right"])) for q in Q]
    uniq_queries = [(h, min(c), max(c)) for h, c in queries]
    assert len(set(uniq_queries)) == n_val, set(uniq_queries)

    # Test the order gets shuffled every iteration
    order = [hash(q) for q in queries]
    round_orders = [order[k * n_val : (k + 1) * n_val] for k in range(n_repeat)]
    for round_order in round_orders:
        assert len(set(round_order)) == n_val
    same_order = [round_orders[0] != order for order in round_orders[1:]]
    assert sum(same_order) in [len(same_order), len(same_order) - 1]


def test_round_robin(server, logs):

    exp = Path(__file__).parent / "data" / "round-robin.yaml"

    with open(exp, "r") as f:
        config = yaml.load(f, Loader=yaml.SafeLoader)
    n = len(config["targets"])
    n_repeat = 4

    with logs:
        server.authorize()
        server.post("/init_exp", data={"exp": exp.read_text()})
        for k in range(n_repeat * n):
            print(k)
            q = server.get("/query").json()

            ans = {"winner": random.choice([q["left"], q["right"]]), "puid": "foo", **q}
            server.post("/answer", json=ans)
            sleep(0.05)

        r = server.get("/responses")
        df = pd.DataFrame(r.json())
        assert set(df["head"].unique()) == set(range(n))
        heads = list(df["head"])
        rounds_heads = [heads[k * n : (k + 1) * n] for k in range(n_repeat)]

        # Make sure every head is asked in every round
        for heads in rounds_heads:
            assert len(set(heads)) == n

        # Make sure round order always shuffled
        orders = [hash(tuple(heads)) for heads in rounds_heads]
        assert len(set(orders)) == len(orders)


def test_round_robin_per_user(server):
    N = 5
    R = 2
    config = {"targets": N, "samplers": {"RoundRobin": {}}}
    server.authorize()
    server.post("/init_exp", data={"exp": config})

    # Ordering in this nested for loop is important
    # (users need to alternately request queries for valid test of per-user RR)
    for k in range(R * N):
        for puid in ["u1", "u2"]:
            q = server.get(f"/query?puid={puid}").json()
            winner = random.choice([q["left"], q["right"]])
            server.post("/answer", json={"winner": winner, "puid": puid, **q})
    r = server.get("/responses")
    df = pd.DataFrame(r.json())
    assert set(df["puid"].unique()) == {"u1", "u2"}
    df["test_user_responses"] = (1 + df["num_responses"]) // 2
    assert (df["puid_num_responses"] == df["test_user_responses"]).all()
    heads = df.pivot(values="head", columns="puid", index="puid_num_responses")

    for puid in heads.columns:
        for r in range(R):
            one_round = heads[puid].iloc[r * N : (r + 1) * N]
            assert set(one_round.values) == set(range(N))



import logging
import random

import numpy as np

from salmon.triplets.samplers._round_robin import (
    RoundRobin, _get_query, _score_query
)
from salmon.triplets.samplers.utils import Answer, Query

logger = logging.getLogger(__name__)


def _random_query(n: int):
    return np.random.choice(n, size=3, replace=False).tolist()


class Validation(RoundRobin):
    """Ask about the same queries repeatedly"""

    def __init__(self, n, d=2, n_queries=20, queries=None, ident=""):
        """
        This sampler asks the same questions repeatedly, useful to evaluate
        query difficulty.

        Parameters
        ----------
        n : int
            Number of objects
        ident : str
            Identifier of the algorithm
        n_queries : int, optional (default=20)
            Number of validation queries.
        d : int
            Embedding dimension.
        queries : List[Tuple[int, int, int]]
            The list of queries to ask about. Each query is
            ``(head, obj1, obj2)`` where ``obj1`` and ``obj2`` are
            randomly shown on the left and right. Each item in the tuple
            is the `index` of the target to ask about. For example:

            .. code-block:: python

                queries=[(0, 1, 2), (3, 4, 5), (6, 7, 8)]

            will first ask about a query with ``head_index=0``, then
            ``head_index=3``, then ``head_index=6``.
        """
        self.n_queries = n_queries
        if queries is None:
            queries = []
            while True:
                q = _random_query(n)
                if q not in queries:
                    queries.append(q)
                if len(queries) == n_queries:
                    break
        idx = [i for query in queries for i in query]
        if n - 1 < max(idx):
            raise ValueError(
                f"The index {max(idx)} is included as an index for validation "
                f"sampling, which is too large for the n={n} targets."
            )
        self._val_queries = queries
        super().__init__(n=n, d=d, ident=ident)

    def get_query(self, **kwargs):
        idx = self.counter % len(self._val_queries)
        if idx == 0:
            random.shuffle(self._val_queries)

        self.counter += 1
        h, l, r = self._val_queries[idx]
        if random.choice([True, False]):
            l, r = r, l
        return {"head": int(h), "left": int(l), "right": int(r)}, float(idx)



import itertools
from copy import deepcopy
from typing import List, Optional, Tuple, Union

import numpy as np
import torch
import torch.optim as optim
from scipy.special import binom
from sklearn.base import BaseEstimator
from skorch.dataset import Dataset as SkorchDataset
from skorch.net import NeuralNet
from skorch.utils import is_dataset
from torch.nn.modules.loss import _Loss
from torch.utils.data import TensorDataset

from salmon.triplets.samplers.adaptive.search import gram_utils, score
from salmon.utils import get_logger

logger = get_logger(__name__)


class Reduce:
    def __call__(self, input: torch.Tensor, target=None) -> torch.Tensor:
        return torch.sum(input)


class NumpyDataset(SkorchDataset):
    def transform(self, X, y):
        return X, torch.from_numpy(np.array([0])) if y is None else y


class Embedding(BaseEstimator):
    """
    An optimization algorithm that produces an embedding from human responses
    of the form ``[head, winner, loser]``.
    """

    def __init__(
        self,
        module,
        module__n: int = 85,
        module__d: int = 2,
        optimizer=optim.Adadelta,
        warm_start=True,
        max_epochs=100,
        initial_batch_size=512,
        **kwargs,
    ):
        """
        Parameters
        ----------
        module : nn.Module
            The noise model to use.
        module__n : int
            The number of items to embed.
        module__d : int, optional (default: 2)
            The number of dimensions to embed into.
        optimizer : torch.optim.Optimizer
            The optimizer to use.
        max_epochs : int, optional (default: 100)
            The number of epochs—or passes through the dataset—to perform.
        warm_start : bool, optional (default: True)
            Whether to use the existing embedding.
        initial_batch_size : int, optional (default: 512)
            The optimizer's (initial) batch size.
        kwargs : dict, optional
            Additional keyword arguments to pass the underlying noise model
            (CKL, TSTE, etc).
        """
        self.module = module
        self.module__n = module__n
        self.module__d = module__d
        self.optimizer = optimizer
        self.warm_start = warm_start
        self.max_epochs = max_epochs
        self.initial_batch_size = initial_batch_size
        self.kwargs = kwargs

    def initialize(self, embedding: Optional[np.ndarray] = None):
        """
        Initialize this optimization algorithm.
        """
        self.meta_ = {"num_answers": 0, "model_updates": 0, "num_grad_comps": 0}
        self.initialized_ = True
        self.answers_ = np.zeros((1000, 3), dtype="uint16")

        self.net_ = NeuralNet(
            module=self.module,
            module__n=self.module__n,
            module__d=self.module__d,
            optimizer=self.optimizer,
            warm_start=self.warm_start,
            **self.kwargs,
            criterion=Reduce,
            verbose=False,
            batch_size=-1,
            max_epochs=self.max_epochs,
            train_split=None,
            dataset=NumpyDataset,
        ).initialize()
        if embedding is not None:
            if not isinstance(embedding, np.ndarray):
                raise ValueError(
                    f"Specify embedding as a NumPy array, not a {type(embedding)}"
                )
            with torch.no_grad():
                em = torch.from_numpy(embedding.astype("float32"))
                self.net_.module_.embedding.data = em
        return self

    # def converged(self):
    #     answers = self.answers_[: self.meta_["num_answers"]]
    #     self.optimizer_.zero_grad()
    #     losses = self.module_.forward(answers)
    #     loss = losses.mean()
    #     loss.backward()
    #     G = self.module_._embedding.grad.detach().numpy().copy()
    #     self.optimizer_.zero_grad()

    #     n, d = G.shape

    #     grad_norms2 = (G ** 2).sum(axis=0)  # Forbenius norm squared
    #     assert grad_norms2.shape == (n, 1)

    #     max_grad_norm2 = grad_norms2.max()
    #     avg_grad_norm2 = grad_norms2.mean()

    #     grad_error = np.sqrt(max_grad_norm2 / avg_grad_norm2)
    #     return grad_error < self.epsilon

    def push(self, answers: Union[list, np.ndarray]):
        """
        Push some answers to train on when ``partial_fit`` is called.

        Parameters
        ----------
        answers : Union[list, np.ndarray]
            Answers with 2 dimensions. The answers should be organized
            ``[head, winner, loser]``.

        Returns
        -------
        nbytes : int
            The number of bytes all of the stored answers.
        """
        if not (hasattr(self, "initialized_") and self.initialized_):
            self.initialize()
        if isinstance(answers, list):
            answers = (
                np.array(answers) if len(answers) else np.empty((0, 3), dtype="uint16")
            )
        num_ans = deepcopy(self.meta_["num_answers"])

        if num_ans + len(answers) >= len(self.answers_):
            n = len(answers) + len(self.answers_)
            new_ans = np.zeros((n, 3), dtype="uint16")
            self.answers_ = np.vstack((self.answers_, new_ans))
        self.answers_[num_ans : num_ans + len(answers)] = answers

        self.meta_["answers_bytes"] = self.answers_.nbytes
        self.meta_["num_answers"] += len(answers)
        return self.answers_.nbytes

    def partial_fit(self, answers, sample_weight=None):
        """
        Process the provided answers.

        Parameters
        ----------
        anwers : array-like
            The answers, with shape (num_ans, 3). Each row is
            [head, winner, loser].

        Returns
        -------
        self

        Notes
        -----
        This function runs one iteration of SGD.

        This impure function modifies

            * ``self.meta_`` keys ``model_updates`` and ``num_grad_comps``
            * ``self.module_``, the embedding.
        """
        if not (hasattr(self, "initialized_") and self.initialized_):
            self.initialize()
        if not isinstance(answers, np.ndarray):
            answers = np.array(answers, dtype="uint16")

        beg_meta = deepcopy(self.meta_)
        eg_deadline = deepcopy(len(answers) + beg_meta["num_grad_comps"])
        if sample_weight is not None:
            sample_weight = torch.from_numpy(sample_weight)
        while True:
            idx_train = self.get_train_idx(self.meta_["num_answers"])
            train_ans = torch.from_numpy(answers[idx_train].astype("int64"))
            losses = self.module_.forward(train_ans)

            if sample_weight is not None:
                losses *= sample_weight[idx_train]

            self.optimizer_.zero_grad()
            loss = losses.mean()
            loss.backward()
            self.optimizer_.step()
            self.optimizer_.zero_grad()
            with torch.no_grad():
                self._project_onto_ball()
            self.optimizer_.zero_grad()

            self.meta_["num_grad_comps"] += len(train_ans)
            self.meta_["model_updates"] += 1
            self.meta_["batch_size"] = len(idx_train)
            logger.info("%s", self.meta_)
            if self.meta_["num_grad_comps"] >= eg_deadline:
                del loss, losses
                break
        return self

    def _project_onto_ball(self):
        norms = torch.norm(self.module_._embedding, dim=1)
        max_norm = 10 * self.module__d
        idx = norms > max_norm
        if idx.sum():
            factor = max_norm / norms[idx]
            d = self.module_._embedding.shape[1]
            if d == 1:
                factor = factor.reshape(-1, 1)
            else:
                factor = torch.stack((factor,) * d).T
            self.module_._embedding[idx] *= factor
        return True

    def score(self, answers, y=None) -> float:
        if not (hasattr(self, "initialized_") and self.initialized_):
            self.initialize()
        with torch.no_grad():
            score = self._score(answers)
        self.meta_["last_score"] = score
        return score

    def _score(self, answers, y=None):
        with torch.no_grad():
            win2, lose2 = self.module_._get_dists(answers)
            acc = (win2 < lose2).numpy().astype("float32").mean().item()
        return acc

    def fit(self, X, y=None):
        if not self.warm_start:
            msg = "Only warm_start=True is accepted, not warm_start={}"
            raise ValueError(msg.format(self.warm_start))
        if not (hasattr(self, "initialized_") and self.initialized_):
            self.initialize()
        for epoch in range(self.max_epochs):
            n_ans = deepcopy(self.meta_["num_answers"])
            self.partial_fit(self.answers_[:n_ans])
        return self

    def embedding(self) -> np.ndarray:
        return self.net_.module_._embedding.detach().numpy()

    @property
    def embedding_(self) -> np.ndarray:
        return self.embedding()

    @property
    def module_(self):
        return self.net_.module_

    @property
    def optimizer_(self):
        return self.net_.optimizer_

    def get_train_idx(self, n_ans):
        bs = min(n_ans, self.initial_batch_size)
        idx = np.random.choice(n_ans, replace=False, size=bs)
        return idx


class GD(Embedding):
    def get_train_idx(self, n_ans):
        return np.arange(n_ans).astype(int)


class OGD(Embedding):
    def __init__(self, dwell=None, initial_batch_size=128, factor=2.0, **kwargs):
        self.dwell = dwell
        self.factor = factor
        super().__init__(initial_batch_size=initial_batch_size, **kwargs)

    def get_train_idx(self, n_ans):
        bs = int(self.initial_batch_size)
        if self.dwell and self.dwell > 0:
            epochs = self.meta_["num_grad_comps"] / n_ans
            n_increases = min(epochs // self.dwell, 100)
            increase_factor = int(self.factor ** n_increases)
            bs = int(bs * increase_factor)

        max_bs = max(5 * self.module__n, n_ans)
        n_idx = min(bs, max_bs)
        return np.random.choice(n_ans, size=min(n_idx, n_ans), replace=False)


class Damper(Embedding):
    """
    Damp the learning rate.
    """

    def __init__(
        self, initial_batch_size=64, max_batch_size=None, **kwargs,
    ):
        self.initial_batch_size = initial_batch_size
        self.max_batch_size = max_batch_size
        super().__init__(**kwargs)

    def get_train_idx(self, len_ans):
        bs = self.batch_size_
        idx_train = np.random.choice(len_ans, size=bs)
        return idx_train

    def initialize(self):
        r = super().initialize()
        if hasattr(self, "max_batch_size"):
            self.max_batch_size_ = self.max_batch_size or 10 * self.module__n
        return r

    def _set_lr(self, lr):
        opt = self.optimizer_
        for group in opt.param_groups:
            group["lr"] = lr

    @property
    def batch_size_(self):
        bs = self.damping()
        self.meta_["batch_size"] = bs
        if self.max_batch_size and bs > self.max_batch_size:
            lr_decay = self.max_batch_size / bs
            if hasattr(self, "optimizer__lr"):
                new_lr = self.optimizer__lr * lr_decay
                self._set_lr(new_lr)
                self.meta_["lr_"] = new_lr
            self.meta_["batch_size"] = self.max_batch_size
        return self.meta_["batch_size"]

    def damping(self):
        raise NotImplementedError


class CntsLRDamper(Damper):
    """
    Decays the learning rate like 1 / (1 + mu) like Thm. 4.7 of [1]_.

    References
    ----------
    1. Bottou, L. A. C., Frank E and Nocedal, Jorge. (2018).
       Optimization methods for large-scale machine learning.
       SIAM Review, 60, 223-223. Retrieved from https://arxiv.org/abs/1606.04838
    """

    def __init__(self, *args, rate=0.05, **kwargs):
        self.rate = rate
        super().__init__(*args, **kwargs)

    def damping(self):
        mu = self.meta_["model_updates"]
        damping = int(self.initial_batch_size * (1 + self.rate * mu))
        self.meta_["damping"] = damping
        return damping


class PadaDampG(Damper):
    def __init__(
        self,
        module,
        module__n=85,
        module__d=2,
        optimizer=None,
        initial_batch_size=64,
        max_batch_size=None,
        growth_factor=1.01,
        dwell=10,
        **kwargs,
    ):
        super().__init__(
            module=module,
            module__n=module__n,
            module__d=module__d,
            optimizer=optimizer,
            initial_batch_size=initial_batch_size,
            max_batch_size=max_batch_size,
            **kwargs,
        )
        self.growth_factor = growth_factor
        self.dwell = dwell

    def damping(self):
        mu = self.meta_["model_updates"]
        if mu % self.dwell == 0:
            pwr = mu / self.dwell
            assert np.allclose(pwr, int(pwr))
            bs = self.initial_batch_size * (self.growth_factor ** (pwr + 1))
            self.meta_["damping"] = int(np.round(bs))
        return self.meta_["damping"]


class GeoDamp(Damper):
    def __init__(
        self,
        module,
        module__n=85,
        module__d=2,
        optimizer=None,
        initial_batch_size=64,
        max_batch_size=None,
        dwell=10,
        growth_factor=1.01,
        **kwargs,
    ):
        super().__init__(
            module=module,
            module__n=module__n,
            module__d=module__d,
            optimizer=optimizer,
            initial_batch_size=initial_batch_size,
            max_batch_size=max_batch_size,
            **kwargs,
        )
        self.growth_factor = growth_factor
        self.dwell = dwell

    def damping(self):
        n_eg = self.meta_["num_answers"]
        if n_eg % self.dwell == 0:
            pwr = n_eg / self.dwell
            assert np.allclose(pwr, int(pwr))
            pwr = int(pwr)
            bs = self.initial_batch_size * (self.growth_factor ** (pwr + 1))
            self.meta_["damping"] = int(np.round(bs))
        return self.meta_["damping"]



import json
import os
import random
from pathlib import Path
from time import sleep, time
from typing import Tuple

import numpy as np
import pandas as pd
import pytest
import yaml

from salmon.triplets.samplers import TSTE

from .utils import LogError, logs, server


def test_backend_basics(server, logs):
    exp = Path(__file__).parent / "data" / "round-robin.yaml"
    exp_config = yaml.safe_load(exp.read_text())

    assert len(exp_config["samplers"]) == 1
    puid = "puid-foo"
    with logs:
        server.authorize()
        server.post("/init_exp", data={"exp": exp.read_text()})
        for k in range(30):
            _start = time()
            q = server.get("/query").json()
            score = max(abs(q["head"] - q["left"]), abs(q["head"] - q["right"]))
            assert q["score"] == score
            ans = {"winner": random.choice([q["left"], q["right"]]), "puid": puid, **q}
            ans["response_time"] = time() - _start
            server.post("/answer", data=ans)
        sleep(5)

    print("Getting responses...")
    r = server.get("/responses")
    print("Done responses...")
    df = pd.DataFrame(r.json())
    assert len(df) == 30


def test_init_errors_propogate(server):
    exp = Path(__file__).parent / "data" / "exp-active-bad.yaml"

    server.authorize()
    server.get("/init")
    r = server.post("/init_exp", data={"exp": exp.read_text()}, error=True)
    assert r.status_code == 500
    assert "module 'salmon.triplets.samplers' has no attribute 'FooBar'" in r.text


def test_run_errors_logged(server, logs):
    # This test is only designed to make sure errors are raised during pytest
    # it's not designed to make sure errors have much detail; the docker logs
    # will reflect more of that.
    config = {
        "targets": list(range(10)),
        "sampling": {"common": {"d": 1}},
        "samplers": {"ARR": {}},
    }
    with pytest.raises(LogError):
        with logs:
            server.authorize()
            server.get("/init")
            r = server.post("/init_exp", data={"exp": config})
            for k in range(10):
                q = server.get("/query").json()
                winner = random.choice([q["left"], q["right"]])
                ans = {"winner": winner, "puid": "", **q}
                ans["left"] = 12
                server.post("/answer", data=ans)
                sleep(2 if k == 3 else 1)
            sleep(5)


def test_backend_random_state():
    n, d = 85, 2
    random_state = 42
    alg1 = TSTE(n=n, d=d, ident="alg1", random_state=random_state)
    alg2 = TSTE(n=n, d=d, ident="alg2", random_state=random_state)
    assert np.allclose(alg1.opt.embedding_, alg2.opt.embedding_)
